#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <ctype.h>
#include "cernlib.h"
#include "histoscope.h"

/* 09/23/2002 igv: added 'F' to HROPEN options in order
 * to avoid using C I/O which in 2002 version of CERNLIB 
 * depends on libshift.so. Changed version number from
 * 1.0 to 1.1.
 */

#define VERSION "1.1"
#define DATE "09/23/2002"
#define AUTHOR "Igor Volobouev, igv@lbl.gov"

#define DEFAULT_HBKMAX 10000000   /* space in words for hbook */
#define DEFAULT_MAXREC 32000
#define MAX_NTUPLE_VARIABLES 1000

int pawc_[DEFAULT_HBKMAX];
int quest_[100];
int hcbook_[51];
int hcbits_[37];
int rzcl_[11];
int lcont, lcid, lcdir;

static int *iq, *lq;
static char chtitl[128];
static int golower = 0;
static float rminmax[2*MAX_NTUPLE_VARIABLES];

static int hist_1d_counter = 0;
static int hist_2d_counter = 0;
static int hist_profile_counter = 0;
static int rwn_counter = 0;
static int cwn_counter = 0;

static void convert_directory(const char *dir, char *category);
static void convert_1d(int id, char *category);
static void convert_2d(int id, char *category);
static void convert_rwn(int id, char *category);
static void convert_cwn(int id, char *category);
static void convert_profile(int id, char *category);

#define check_memory(ptr) do {\
    if (ptr == NULL)\
    {\
	fprintf(stderr, "Fatal error: out of memory. Exiting.\n");\
	exit(EXIT_FAILURE);\
    }\
} while(0);

#ifdef DEBUG_PRINT
#undef DEBUG_PRINT
#endif

int main(int argc, char **argv)
{
    int printhelp = 0, printversion = 0, hlim = DEFAULT_HBKMAX, lun = 25;
    int ierr = 0, record_size = 0, maxrec = DEFAULT_MAXREC, showdir = 0;
    long tmplong;
    char *progname, *endptr, *category = "", *infile = NULL, *ofile = NULL;
    /* The topdir[] array must start with // and must be at most 10 characters long */
    char topdir[] = "//TOPDIR";
    char hropen_opt[] = "PQF";
    char *ctmp;

    /* Variables needed for 'getopt' */
    int c;
    int errflg = 0;
    extern char *optarg;
    extern int optind;

    progname = strrchr(argv[0], '/');
    if (progname)
        ++progname;
    else
        progname = argv[0];

    while((c = getopt(argc, argv, "c:dehlm:o:r:v")) >= 0)
    {
        switch(c)
        {
        case 'c':
            category = optarg;
            break;
	case 'd':
	    showdir = 1;
	    break;
        case 'h':
            printhelp = 1;
            break;
	case 'l':
	    golower = 1;
	    break;
	case 'm':
	    tmplong = strtol(optarg, &endptr, 0);
	    if (optarg == endptr)
		++errflg;
	    else if (tmplong < 0 || tmplong > 65000)
	    {
		fprintf(stderr, "Argument \"%s\" is out of range\n", optarg);
		++errflg;
	    }
	    else
		maxrec = tmplong;
	    break;
        case 'o':
            ofile = optarg;
            break;
	case 'r':
	    tmplong = strtol(optarg, &endptr, 0);
	    if (optarg == endptr)
		++errflg;
	    else if (tmplong < 0 || tmplong > INT_MAX)
	    {
		fprintf(stderr, "Argument \"%s\" is out of range\n", optarg);
		++errflg;
	    }
	    else
		record_size = tmplong;
	    break;
        case 'v':
            printversion = 1;
            break;
        default:
            ++errflg;
            break;
	}
    }
    if (optind != argc-1)
	++errflg;
    else
	infile = argv[optind];

    if (printversion)
    {
	printf("%s version %s  HBOOK to Histo-Scope Converter\n",
	       progname, VERSION);
	printf("Author: %s\n", AUTHOR);
	printf("Date: %s\n", DATE);
	if (argc == 2)
	    exit(EXIT_SUCCESS);
    }
    if (printhelp || errflg)
    {
	/* Print the usage instructions */
	printf("\nUsage: %s [-c category] [-d] [-h] [-l] [-m maxrec] [-o ofile] [-r reclen] "
	       "\n       [-v] filename.hbook\n\n", progname);
	printf("The following command line switches are supported:\n\n");
	printf("  -c    category is the name of the Histo-Scope category which will be\n"
	       "        prepended to all HBOOK directory names.\n\n");
	printf("  -d    show the HBOOK file subdirectory tree and exit.\n\n");
	printf("  -h    print program usage info and exit.\n\n");
	printf("  -l    convert ntuple variable names to lower case.\n\n");
	printf("  -m    maxrec is the maximun number of records in the file (can not exceed"
	       "\n        65000). Default is %d.\n\n", DEFAULT_MAXREC);
	printf("  -o    ofile is the name of the output file. If this option is not specified,\n"
	       "        the output file name will be the same as the input file name but with\n"
	       "        extension \".hs\".\n\n");
	printf("  -r    reclen is the HBOOK record length. Must be specified if > 8092.\n\n");
	printf("  -v    print program version.\n");
	printf("\n");

	if (printhelp)
	    exit(EXIT_FAILURE);
	else
	    exit(EXIT_SUCCESS);
    }

    /* Some preliminary initialization */
    lq = &pawc_[9];
    iq = &pawc_[17];

    /* Tell HBOOK how much memory it has */
    hlimit_(&hlim);

    /* Open the HBOOK file */
    quest_[9] = maxrec;
    hropen_(&lun, topdir+2, infile, hropen_opt, &record_size,
	    &ierr, strlen(topdir)-2, strlen(infile), strlen(hropen_opt));
    if (ierr)
    {
	fprintf(stderr, "Failed to open file \"%s\".\n", infile);
	fprintf(stderr, "HROPEN error was %d\n", ierr);
	exit(EXIT_FAILURE);
    }
    if (showdir)
    {
	/* Just show the contents and exit */
	hldir_(topdir, "T", strlen(topdir), 1);
	exit(EXIT_SUCCESS);
    }

    /* Initialize Histo-Scope */
    hs_initialize("HBOOK converter");

    convert_directory(topdir, category);

    /* Figure out the new file name */
    if (ofile == 0)
    {
	ofile = (char *)malloc(strlen(infile) + strlen(".hs") + 1);
	check_memory(ofile);
	ctmp = strrchr(infile, '/');
	if (ctmp)
	    strcpy(ofile, ctmp+1);
	else
	    strcpy(ofile, infile);
	ctmp = strrchr(ofile, '.');
	if (ctmp)
	    *ctmp = '\0';
	strcat(ofile, ".hs");
    }

    /* Write the stuff out */
    ierr = hs_save_file(ofile);
    if (ierr > 0)
    {
	printf("File %s contains %d 1-d histogram(s), %d 2-d histogram(s), "
	       "and %d ntuple(s).\n", ofile, hist_1d_counter,
	       hist_2d_counter, rwn_counter);
	if (hist_profile_counter + cwn_counter == 0)
	    printf("All items from the input file have been successfully converted.\n");
	else
	{
	    if (hist_profile_counter > 0)
		printf("%d profile histograms have been skipped.\n", hist_profile_counter);
	    if (cwn_counter > 0)
		printf("%d column-wise ntuples have been skipped.\n", cwn_counter);
	}
	exit(EXIT_SUCCESS);
    }
    else
	exit(EXIT_FAILURE);
}

static void convert_directory(const char *dir, char *category)
{
    /* Based on h2root "convert_directory" function */
    int id, key, z0, i999, ls, ndir, nch, ncw, i, k, lcateg;
    const int KLS = 26, KNSD = 23;
    char chdir[17], hbookdir[17];
    int ihdir[4];
    char *categ = NULL;

#ifdef DEBUG_PRINT
    printf("Converting directory %s\n", dir);
#endif

    for (key=1; key<1000000; ++key)
    {
	i = key;
	z0 = 0;
	rzink_(&i, &z0, "S", 1);
	if (quest_[0]) break;
	if (quest_[13] & 8) {
	    continue;
	}
	id = quest_[20];
	i999 = 999;
	i = 0;
	hrin_(&id, &i999, &i);
	if (quest_[0]) {
	    printf("Error: failed to read HBOOK item with id %d.\n", id);
	    break;
	}
	hdcofl_();
	lcid  = hcbook_[10];
	lcont = lq[lcid-1];
	if (hcbits_[3]) {
	    if (iq[lcid-2] == 2)
		convert_rwn(id, category);
	    else
		convert_cwn(id, category);
	    hdelet_(&id);
	    continue;
	}
	if (hcbits_[0] && hcbits_[7]) {
	    convert_profile(id, category);
	    hdelet_(&id);
	    continue;
	}
	if (hcbits_[0]) {
	    convert_1d(id, category);
	    hdelet_(&id);
	    continue;
	}
	if (hcbits_[1] || hcbits_[2]) {
	    convert_2d(id, category);
	    hdelet_(&id);
	}
    }

    /* converting subdirectories of this directory */
    lcdir = rzcl_[2];
    ls = iq[lcdir+KLS];
    ndir = iq[lcdir+KNSD];
    lcateg = strlen(category);
    for (k=0; k<ndir; ++k) {
	lcdir = rzcl_[2];
	ncw = 4;
	zitoh_(&iq[lcdir+ls+7*k], ihdir, &ncw);
	for (i=0; i<17; ++i)
	    chdir[i] = '\0';
	nch = 16;
	uhtoc_(ihdir, &ncw, chdir, &nch, 16);
	strcpy(hbookdir, chdir);
	for (i=16; i>0; --i) {
	    if (chdir[i] == 0) continue;
	    if (chdir[i] != ' ') break;
	    chdir[i] = '\0';
	}
	categ = (char *)malloc(lcateg + strlen(chdir) + 2);
	strcpy(categ, category);
	if (lcateg > 0)
	    strcat(categ, "/");
	strcat(categ, chdir);
	hcdir_(hbookdir, " ", 16, 1);
	convert_directory(chdir, categ);
	hcdir_("\\", " ", 1, 1);
	free(categ);
    }
}

static void convert_1d(int id, char *category)
{
    int ncx, ncy, nwt, idb, hs_id, nbins, i;
    float xmin, xmax, ymin, ymax, ofl, ufl;
    static float *data = NULL;
    static int oldbins = -1;

#ifdef DEBUG_PRINT
    printf("Converting 1d histo with id %d\n", id);
#endif
    ++hist_1d_counter;

    hgive_(&id, chtitl, &ncx, &xmin, &xmax, &ncy, &ymin, &ymax, &nwt, &idb, 80);
    chtitl[4*nwt] = 0;
    hs_id = hs_create_1d_hist(id, chtitl, category, "X", "Y", ncx, xmin, xmax);
    if (hs_id <= 0)
    {
	fprintf(stderr, "Error: failed to convert 1d histogram with id %d\n", id);
	return;
    }
    nbins = ncx;
    if (nbins > oldbins)
    {
	data = (float *)realloc(data, nbins*sizeof(float));
	check_memory(data);
	oldbins = nbins;
    }
    i = 0;
    hunpak_(&id, data, " ", &i, 1);
    hs_1d_hist_block_fill(hs_id, data, NULL, NULL);

    /* Take care of errors */
    if (hcbits_[8])
    {
	i = 0;
	hunpke_(&id, data, " ", &i, 1);
	hs_set_1d_errors(hs_id, data, NULL);
    }

    /* Take care of under- and overflows */
    i = 0; ufl = hi_(&id, &i);
    i = ncx + 1; ofl = hi_(&id, &i);
    hs_1d_hist_set_overflows(hs_id, ufl, ofl);
}

static void convert_2d(int id, char *category)
{
    int ncx, ncy, nwt, idb, hs_id, nbins, i, j;
    float xmin, xmax, ymin, ymax, ofl[3][3];
    float *transpose;
    static float *data = NULL;
    static int oldbins = -1;

#ifdef DEBUG_PRINT
    printf("Converting 2d histo with id %d\n", id);
#endif
    ++hist_2d_counter;

    hgive_(&id, chtitl, &ncx, &xmin, &xmax, &ncy, &ymin, &ymax, &nwt, &idb, 80);
    chtitl[4*nwt] = '\0';
    hs_id = hs_create_2d_hist(id, chtitl, category, "X", "Y", "Z",
			      ncx, ncy, xmin, xmax, ymin, ymax);
    if (hs_id <= 0)
    {
	fprintf(stderr, "Error: failed to convert 2d histogram with id %d\n", id);
	return;
    }
    nbins = ncx*ncy;
    if (nbins > oldbins)
    {
	data = (float *)realloc(data, 2*nbins*sizeof(float));
	check_memory(data);
	oldbins = nbins;
    }
    transpose = data + nbins;
    i = 0;
    hunpak_(&id, data, " ", &i, 1);
    for (i=0; i<ncx; ++i)
	for (j=0; j<ncy; ++j)
	    transpose[j+i*ncy] = data[i+j*ncx];
    hs_2d_hist_block_fill(hs_id, transpose, NULL, NULL);

    /* Take care of errors */
    if (lq[lcont])
    {
	i = 0;
	hunpke_(&id, data, " ", &i, 1);
	for (i=0; i<ncx; ++i)
	    for (j=0; j<ncy; ++j)
		transpose[j+i*ncy] = data[i+j*ncx];
	hs_set_2d_errors(hs_id, transpose, NULL);
    }

    /* Take care of overflows */
    i = 0; j = 0; ofl[0][2] = hij_(&id, &i, &j);
    i = 0; j = 1; ofl[0][1] = hij_(&id, &i, &j);
    i = 0; j = ncy+1; ofl[0][0] = hij_(&id, &i, &j);
    i = 1; j = 0; ofl[1][2] = hij_(&id, &i, &j);
    ofl[1][1] = 0.f;
    i = 1; j = ncy+1; ofl[1][0] = hij_(&id, &i, &j);
    i = ncx+1; j = 0; ofl[2][2] = hij_(&id, &i, &j);
    i = ncx+1; j = 1; ofl[2][1] = hij_(&id, &i, &j);
    i = ncx+1; j = ncy+1; ofl[2][0] = hij_(&id, &i, &j);
    hs_2d_hist_set_overflows(hs_id, ofl);
}

static void convert_rwn(int id, char *category)
{
    const int Nchar = 9;
    int i, j, ierr = 0, nentries, nvar, first, last, hs_id;
    float *rmin = rminmax, *rmax = rminmax+MAX_NTUPLE_VARIABLES, *x = NULL;
    char *chtag_out, *name;
    char **varnames;

#ifdef DEBUG_PRINT
    printf("Converting rwn with id %d\n", id);
#endif
    ++rwn_counter;

    hgiven_(&id, chtitl, &nvar, "", rmin, rmax, 80, 0);
    rmin = (float *)malloc(3*nvar*sizeof(float));
    check_memory(rmin);
    rmax = rmin+nvar;
    x = rmax+nvar;
    varnames = (char **)malloc(nvar*sizeof(char *));
    check_memory(varnames);
    chtag_out = (char *)malloc((nvar*Nchar+1)*sizeof(char));
    check_memory(chtag_out);
    chtag_out[nvar*Nchar] = '\0';
    hgiven_(&id, chtitl, &nvar, chtag_out, rmin, rmax, 80, Nchar);
    chtitl[81] = '\0';
    for (i=80; i>0; --i) {
	if (chtitl[i] == ' ')
	    chtitl[i] = '\0';
	else
	    break;
    }
    name = chtag_out;
    for(i=0; i<nvar; ++i) {
	name[Nchar-1] = '\0';
	first = last = 0;

	/* suppress traling blanks */
	for (j=Nchar-2; j>0; --j) {
	    if (golower) name[j] = tolower(name[j]);
	    if (name[j] == ' ' && last == 0)
		name[j] = '\0';
	    else
		last = j;
	}
	if (golower)
	    name[0] = tolower(name[0]);

	/* suppress leading blanks */
	for (j=0; j<Nchar; ++j) {
	    if (name[j] != ' ') break;
	    first = j+1;
	}
	varnames[i] = name+first;
	name += Nchar;
    }

    hs_id = hs_create_ntuple(id, chtitl, category, nvar, varnames);
    if (hs_id <= 0)
    {
	fprintf(stderr, "Error: failed to convert an ntuple with id %d\n", id);
    }
    else
    {
	/* Cycle over the ntuple */
	hnoent_(&id, &nentries);
	hgnpar_(&id, "convert_rwn", strlen("convert_rwn"));
	for (i=1; i<=nentries; ++i) {
	    hgnf_(&id, &i, x, &ierr);
	    if (hs_fill_ntuple(hs_id, x) < 0)
	    {
		fprintf(stderr, "Fatal error: out of memory. Exiting.\n");
		exit(EXIT_FAILURE);
	    }
	}
    }

    free(chtag_out);
    free(varnames);
    free(rmin);
}

static void convert_cwn(int id, char *category)
{
    static int warned = 0;
    ++cwn_counter;
    if (!warned)
    {
	printf("WARNING! Conversion of column-wise ntuples is not supported!\n");
	warned = 1;
    }
    printf("Skipped column-wise ntuple with id %d\n", id);
}

static void convert_profile(int id, char *category)
{
    static int warned = 0;
    ++hist_profile_counter;
    if (!warned)
    {
	printf("WARNING! Conversion of profile histograms is not supported!\n");
	warned = 1;
    }
    printf("Skipped profile histogram with id %d\n", id);
}
