#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <float.h>
#include "cernlib.h"
#include "dd.h"
#include "sbshape.h"

/* Couple functions from johnson.f file */
void sbfitlargegamma_(const double *MEAN,const double *SIGMA,
                      const double *SKEW,const double *KURT,
                      double *GAMMA,double *DELTA,double *XLAM,
                      double *XI,int *FAULT);

void sbfit_(const double *MEAN,const double *SIGMA,
            const double *SKEW,const double *KURT,
            double *GAMMA,double *DELTA,double *XLAM,
            double *XI,int *FAULT);

#define TAILTERMS 180        /* Do not increase beyond 184 */
#define GAUSS_QUAD_POINTS 96 /* Should be one of the numbers allowed
                                by DGQUAD (CERNLIB entry D107) */
#define SMALL_A_BOUND 1.0
#define LARGE_RATIO_BOUND 5.5

/* For the standard algorithm, LARGE_RATIO_MIN_A should be set 
 * around 0.231. This avoids problem for a = 0.230, b = 27.2.
 * However, it appears that the setting below is sufficient
 * because the "death" region is covered by the large gamma
 * approximation.
 */
#define LARGE_RATIO_MIN_A 0.122

#define SQRPI 1.7724538509055160273
#define SQR2 1.4142135623730950488

static int initialized = 0;
static int pos_tail_coeff[6*TAILTERMS];
static int neg_tail_coeff[6*TAILTERMS];
static double max_exponent = 0.0;
static double gauss_abscissae[GAUSS_QUAD_POINTS];
static double gauss_weights[GAUSS_QUAD_POINTS];
static double gauss_data[6*GAUSS_QUAD_POINTS];

static void john_init(void)
{
    int k, j, ngpoints;
    double dmin, dmax;

    /* Fill the coefficients for the tail series */
    for (k=0; k<TAILTERMS; ++k)
    {
	pos_tail_coeff[0*TAILTERMS+k] = 1;
	pos_tail_coeff[1*TAILTERMS+k] = k+1;
	pos_tail_coeff[2*TAILTERMS+k] = ((k+1)*(k+2))/2;
	pos_tail_coeff[3*TAILTERMS+k] = ((k+1)*(k+2)*(k+3))/6;
	pos_tail_coeff[4*TAILTERMS+k] = ((k+1)*(k+2)*(k+3)*(k+4))/24;
        /* Need to make sure that the next coefficient
	   is calculated without integer overflow */
	if (k % 5 == 0)
	    pos_tail_coeff[5*TAILTERMS+k] = 
                pos_tail_coeff[4*TAILTERMS+k] * ((k+5)/5);
	else
	    pos_tail_coeff[5*TAILTERMS+k] = 
                (pos_tail_coeff[4*TAILTERMS+k]/5) * (k+5);
        neg_tail_coeff[0*TAILTERMS+k] = -1;
	neg_tail_coeff[1*TAILTERMS+k] = k-1;
	neg_tail_coeff[2*TAILTERMS+k] = -((k-1)*(k-2))/2;
	neg_tail_coeff[3*TAILTERMS+k] = ((k-1)*(k-2)*(k-3))/6;
	neg_tail_coeff[4*TAILTERMS+k] = -((k-1)*(k-2)*(k-3)*(k-4))/24;
        /* Need to make sure that the next coefficient
	   is calculated without integer overflow */
	if (k % 5 == 0)
	    neg_tail_coeff[5*TAILTERMS+k] = 
                -neg_tail_coeff[4*TAILTERMS+k] * ((k-5)/5);
	else
	    neg_tail_coeff[5*TAILTERMS+k] = 
                -(neg_tail_coeff[4*TAILTERMS+k]/5) * (k-5);
	if (k%2)
	    for (j=0; j<6; ++j)
	    {
		pos_tail_coeff[j*TAILTERMS+k] *= -1;
		neg_tail_coeff[j*TAILTERMS+k] *= -1;
	    }
    }
    for (j=0; j<6; ++j)
	neg_tail_coeff[j*TAILTERMS] = 0;

    /* Get abscissae and weights for gaussian quadrature */
    dmin = -1.0;
    dmax = 1.0;
    ngpoints = GAUSS_QUAD_POINTS;
    dgset_(&dmin, &dmax, &ngpoints, gauss_abscissae, gauss_weights);

    /* Max exponent (for practical purposes) */
    max_exponent = log(DBL_MAX/2.0) - 0.1;

    initialized = 1;
}

static double poly_sum(const double *a, int deg, const double x)
{
    /* Dimension of a should be at least deg+1 */
    register double sum = 0.0;
    for (; deg>=0; --deg)
    {
	sum *= x;
	sum += a[deg];
    }
    return sum;
}

/* The following function returns erfc(x)/exp(-x^2) for non-negative x */
static double erfc_exp_factor(double x)
{
    static const double coeffs[] = {
        1.0, -1.0, 3.0, -15.0, 105.0, -945.0, 10395.0,
        -135135.0, 2027025.0, -34459425.0,
        6.54729075e8, -1.3749310575e10, 3.16234143225e11,
        -7.905853580625e12, 2.13458046676875e14,
        -6.190283353629375e15, 1.91898783962510625e17};
    static const int degree = sizeof(coeffs)/sizeof(coeffs[0])-1;

    assert(x >= 0.0);
    if (x >= 9.0)
        return poly_sum(coeffs, degree, 0.5/x/x)/x/SQRPI;
    else
        return erfc(x)/exp(-x*x);
}

static dd_t kahan_sum(double *arr, int n)
{
    double s, t, sum = 0.0, r = 0.0;
    for (--n; n>=0; --n)
    {
	s = arr[n] - r;
	t = sum + s;
	r = (t - sum) - s;
	sum = t;
    }
    return dd_real(sum, r);
}

static void zeroth_moment_partial_over_b(const double a,
                                         const double b,
                                         const dd_t mom[6],
                                         dd_t deriv[4])
{
    int i;
    for (i=0; i<4; ++i)
        deriv[i] = dd_times_d(dd_diff(mom[i+1], mom[i]), (double)(i+1)/a);
}

static void zeroth_moment_partial_over_a(const double a,
                                         const double b,
                                         const dd_t mom[6],
                                         dd_t deriv[4])
{
    int i;
    for (i=0; i<4; ++i)
    {
        const int n = i+1;
        const dd_t twoab   = dd_real(2.0*a*b, 0.0);
        const dd_t minusc0 = dd_minus_d(twoab, (double)n);
        const dd_t c1      = dd_minus_d(twoab, (double)(2*n+1));
        deriv[i] = dd_times_d(dd_diff(dd_sum(dd_times_d(mom[i+2],(double)(n+1)),
                                             dd_prod(mom[i+1],c1)),
                                      dd_prod(mom[i],minusc0)),
                              -(double)n/2.0/a/a/a);
    }
}

#define convert_zeroth_partials do {\
    meansquared = dd_square(mom[0]);\
    m1d0     = dd_prod(mom[1],der0[0]);\
    deriv[0] = dd_real(0.0, 0.0);\
    deriv[1] = dd_diff(der0[1], dd_times_d(dd_prod(mom[0],der0[0]),2.0));\
    deriv[2] = dd_diff(dd_sum(der0[2],dd_times_d(dd_prod(meansquared,der0[0]),6.0)),\
                       dd_times_d(dd_sum(m1d0, dd_prod(mom[0],der0[1])),3.0));\
    deriv[3] = dd_sum(dd_sum(der0[3], dd_times_d(dd_prod(meansquared,deriv[1]),6.0)),\
                      dd_times_d(dd_diff(dd_prod(mom[0],\
                                                 dd_diff(dd_times_d(m1d0,3.0),der0[2])),\
                                         dd_prod(mom[2],der0[0])),4.0));\
} while(0);

static void central_moment_partial_over_b(const double a,
                                          const double b,
                                          const dd_t mom[6],
                                          dd_t deriv[4])
{
    dd_t meansquared, m1d0, der0[4];

    zeroth_moment_partial_over_b(a, b, mom, der0);
    convert_zeroth_partials;
}

static void central_moment_partial_over_a(const double a,
                                          const double b,
                                          const dd_t mom[6],
                                          dd_t deriv[4])
{
    dd_t meansquared, m1d0, der0[4];

    zeroth_moment_partial_over_a(a, b, mom, der0);
    convert_zeroth_partials;
}

static void central_integ_small_a(const double a, const double b, dd_t *p)
{
    double expo, f;
    int i, j, prec;

    for (i=0; i<GAUSS_QUAD_POINTS; ++i)
    {
	f = 1.0/(1.0+exp(-gauss_abscissae[i]));
	expo = (gauss_abscissae[i]*a + b);
	expo = exp(-expo*expo)*gauss_weights[i];
	for (j=0; j<6; ++j)
	{
	    expo *= f;
	    /* Avoid NaNs */
	    if (expo != 0.0)
		gauss_data[j*GAUSS_QUAD_POINTS+i] = expo;
	    else
		gauss_data[j*GAUSS_QUAD_POINTS+i] = 0.0;
	}
    }
    prec = dd_getprecision();
    dd_setprecision(DD_DBLPREC);
    for (j=0; j<6; ++j)
	p[j] = dd_times_d(kahan_sum(gauss_data+j*GAUSS_QUAD_POINTS,
				    GAUSS_QUAD_POINTS), a/SQRPI);
    dd_setprecision(prec);
}

static void central_integ_big_a(const double a, const double b, dd_t *p)
{
    double expo, f, x, dmin, dmax, halfrange, dtmp;
    volatile double z;
    int i, j, prec;

    dmin = (erfc(a-b))/2.0;
    dmax = (erfc(-a-b))/2.0;
    halfrange = (dmax - dmin)/2.0;
    for (i=0; i<GAUSS_QUAD_POINTS; ++i)
    {
	z = dmin + (gauss_abscissae[i] + 1.0)*halfrange;
	if (z > 0.0 && z < 1.0)
	{
            dtmp = z;
	    x = dgausn_(&dtmp)/SQR2;
	    f = 1.0/(1.0+exp(-(x-b)/a));
	}
	else if (z <= 0.0)
	    f = 0.0;
	else
	    f = 1.0;
	expo = gauss_weights[i];
	for (j=0; j<6; ++j)
	{
	    expo *= f;
	    /* Avoid NaNs */
	    if (expo != 0.0)
		gauss_data[j*GAUSS_QUAD_POINTS+i] = expo;
	    else
		gauss_data[j*GAUSS_QUAD_POINTS+i] = 0.0;
	}
    }
    prec = dd_getprecision();
    dd_setprecision(DD_DBLPREC);
    for (j=0; j<6; ++j)
	p[j] = dd_times_d(kahan_sum(gauss_data+j*GAUSS_QUAD_POINTS,
				    GAUSS_QUAD_POINTS), halfrange);
    dd_setprecision(prec);
}

static void poly_and_deriv(const long double *a, size_t deg,
                                  const long double x,
                                  long double *value, long double *deriv)
{
    /* Dimension of a should be at least deg+1 */
    register long double sum = 0.0L, der = 0.0L;
    for (; deg>=1; --deg)
    {
	sum *= x;
        der *= x;
	sum += a[deg];
        der += deg*a[deg];
    }
    *value = sum*x + a[0];
    *deriv = der;
}

static long double poly_sum_ld(const long double *a, int deg,
                                      const long double x)
{
    /* Dimension of a should be at least deg+1 */
    register long double sum = 0.0L;
    for (; deg>=0; --deg)
    {
	sum *= x;
	sum += a[deg];
    }
    return sum;
}

static void big_a_shape_natural_pars(const double o, const long double f0,
                                     size_t maxdeg, int *ifail,
                                     double *mean, double *var,
                                     double *skew, double *kurt,
                                     double *dskewdo, double *dskewdf0,
                                     double *dkurtdo, double *dkurtdf0)
{
    static const long double meancoeffs[19][37] = {
        {0L,0.5L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L},{0L,0.125L,-0.1875L,0.0625L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0L,0.015625L,-0.1171875L,0.1953125L,-0.1171875L,0.0234375L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0L,0.00130208333333333333333333333333L,-0.041015625L,
         0.195963541666666666666666666667L,-0.341796875L,0.2734375L,-0.1025390625L,
         0.0146484375L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0L,0.0000813802083333333333333333333333L,-0.0103759765625L,
         0.123087565104166666666666666667L,-0.4742431640625L,0.8485107421875L,
         -0.8074951171875L,0.4229736328125L,-0.1153564453125L,0.0128173828125L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0L,4.06901041666666666666666666667e-6L,-0.002081298828125L,
         0.0579854329427083333333333333333L,-0.444793701171875L,1.50592041015625L,
         -2.7387542724609375L,2.9290924072265625L,-1.90338134765625L,
         0.740203857421875L,-0.1586151123046875L,0.0144195556640625L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0L,1.69542100694444444444444444444e-7L,-0.000347137451171875L,
         0.0221782260470920138888888888889L,-0.322027842203776041666666666667L,
         1.90951054890950520833333333333L,-5.92633056640625L,10.90130615234375L,
         -12.681278228759765625L,9.59973907470703125L,-4.7254085540771484375L,
         1.4605808258056640625L,-0.2577495574951171875L,0.0198268890380859375L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0L,6.0550750248015873015873015873e-9L,
         -0.0000496001470656622023809523809524L,0.00719070737324063740079365079365L,
         -0.19235134124755859375L,1.91431427001953125L,-9.552484989166259765625L,
         27.843293666839599609375L,-51.648101806640625L,64.0187168121337890625L,
         -54.34219837188720703125L,31.74615383148193359375L,
         -12.565290927886962890625L,3.22186946868896484375L,
         -0.4832804203033447265625L,0.0322186946868896484375L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{0L,1.89221094525049603174603174603e-10L,
         -6.20030221484956287202380952381e-6L,0.00203013656631348625062003968254L,
         -0.0985374464875175839378720238095L,1.60442978172075180780319940476L,
         -12.4217143356800079345703125L,54.7258019149303436279296875L,
         -152.1110164225101470947265625L,283.9810342490673065185546875L,
         -369.9214361608028411865234375L,343.3636428415775299072265625L,
         -228.7406502664089202880859375L,108.6535204946994781494140625L,
         -35.9439812600612640380859375L,7.8734435141086578369140625L,
         -1.0269708931446075439453125L,0.0604100525379180908203125L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{0L,5.25614151458471119929453262787e-12L,
         -6.88930352528889973958333333333e-7L,0.000508395297291838092568479938272L,
         -0.0443868021722193117494936342593L,1.16362524970814033790870949074L,
         -13.6610061753955152299669053819L,88.279010452389244049314468626L,
         -353.85104589164257049560546875L,947.560392158726851145426432292L,
         -1778.07000944390892982482910156L,2410.50902364775538444519042969L,
         -2404.08316534012556076049804687L,1777.9039643704891204833984375L,
         -973.183292616158723831176757813L,389.219115804880857467651367188L,
         -110.5705328285694122314453125L,21.138484217226505279541015625L,
         -2.4390558712184429168701171875L,0.1283713616430759429931640625L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{0L,1.31403537864617779982363315697e-13L,
         -6.88932323581957943225033068783e-8L,0.000114475058250431451006961667769L,
         -0.017888200469315052032470703125L,0.747277971692383289337158203125L,
         -13.1004462673018376032511393229L,121.679448559507727622985839844L,
         -685.613888766430318355560302734L,2551.25422579555047882927788628L,
         -6629.81833250029012560844421387L,12492.2262919140048325061798096L,
         -17500.35731400316581130027771L,18515.6260640001855790615081787L,
         -14913.1328637013211846351623535L,9150.62849590089172124862670898L,
         -4247.29059227742254734039306641L,1466.31973884534090757369995117L,
         -364.94373473105952143669128418L,61.8910427321679890155792236328L,
         -6.40252166194841265678405761719L,0.304881983902305364608764648437L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{0L,2.98644404237767681778098444765e-15L,
         -6.26302560313840850492461820587e-9L,0.000023423181601849450158331282192L,
         -0.00653212043092088409201808707424L,0.431915461953000101462873832259L,
         -11.1876284283353015780448913574L,147.245794741553254425525665283L,
         -1145.56851400528103113174438477L,5794.92838648613542318344116211L,
         -20305.7981716669164597988128662L,51479.3983253301121294498443604L,
         -97332.683711075806058943271637L,140146.315217037918046116828918L,
         -155795.00797190034063532948494L,134743.599291388120036572217941L,
         -90858.3407332794740796089172363L,47592.9279691271949559450149536L,
         -19165.6284689594758674502372742L,5819.14522985054645687341690063L,
         -1288.50748446711804717779159546L,196.343997633084654808044433594L,
         -18.4072497781016863882541656494L,0.800315207743551582098007202148L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{0L,6.22175842162016003704371759927e-17L,
         -5.21918893587910366379452072811e-10L,4.39249894887820761434781252557e-6L,
         -0.00218249749366858269597009643829L,0.227406978877632999970388465993L,
         -8.63885331720693675042302520187L,159.470850354288202690739164907L,
         -1690.92206705764692742377519608L,11444.7770705024546865994731585L,
         -53055.5505721531517338007688522L,176906.289460123080061748623848L,
         -439301.245672291988739743828773L,832988.541105764015810564160347L,
         -1.22765571974613703787326812744e6L,1.4233407041923608630895614624e6L,
         -1.3076429162810843990882858634e6L,954699.460294725577114149928093L,
         -553060.97634627531078876927495L,252577.597870370482269208878279L,
         -89792.8653238022889127023518085L,24326.970175489805114921182394L,
         -4851.07728527054860023781657219L,671.097648159957316238433122635L,
         -57.5226555565677699632942676544L,2.30090622226271079853177070618L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L},{0L,1.19649200415772308404686876909e-18L,
         -4.0147608993808034419619400902e-11L,7.60290387200056093712890592129e-7L,
         -0.000672424652318055474948802781025L,0.110086011621609733123269546714L,
         -6.10121776588563410438258539547L,156.78254805909182139639413417L,
         -2243.88345952154243908201654752L,20081.5814456842087599431612977L,
         -121433.078729813705649576149881L,523730.794593024593268637545407L,
         -1.67494331988941976305795833468e6L,4.08669534326734574278816580772e6L,
         -7.76812997337545830305316485465e6L,1.16801297725396014539001043886e7L,
         -1.40410620303721865639090538025e7L,1.35870445418254543255898170173e7L,
         -1.06179723861728422889427747577e7L,6.69969858975115248540532775223e6L,
         -3.39968960804037919842812698334e6L,1.3757334697660603524127509445e6L,
         -437688.448063535247456457000226L,107039.595526050220541947055608L,
         -19413.8962503416223626118153334L,2459.09352504327216593082994223L,
         -194.138962503416223626118153334L,7.1903319445709712454117834568L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L},{0L,2.13659286456736265008369423052e-20L,
         -2.86768638874946685562611124425e-12L,1.22193111122251288028396290852e-7L,
         -0.00019226388475671252997994431209L,0.0493617181277344629785553260105L,
         -3.9763315538346193874292451926L,141.448535586646447348083177256L,
         -2712.56585658969349790456728986L,31810.4844289413050162950483612L,
         -248226.303372477824604218421016L,1.36738812490918409956802428323e6L,
         -5.54969987937792738989628560375e6L,1.71281451288779322601385501912e7L,
         -4.11623445892981495717322104611e7L,7.84090039559873677490031695925e7L,
         -1.19963729919516940344692557119e8L,1.48826857767110529289311671164e8L,
         -1.50663589685328254461182950763e8L,1.24889016124024460452801577048e8L,
         -8.48205501738702087664023565594e7L,4.70900458123169209443403815385e7L,
         -2.12491478447565557985399209429e7L,7.71744542111983008680908824317e6L,
         -2.22151596967971656226836785208e6L,495067.296981211597994843032211L,
         -82339.1874717614058454273617826L,9617.96776735674541214393684641L,
         -703.753739074883810644678305835L,24.2673703129270279532647691667L,0L,0L,0L,0L,
         0L,0L,0L},{0L,3.56098810761227108347282371753e-22L,
         -1.91179093117446006516914745471e-13L,1.83292056422036003759027930839e-8L,
         -0.0000512917531345952283554601239656L,0.0206250678283052627792897705511L,
         -2.408439075099712987366883302L,118.10708978718640141063592355L,
         -3017.86690702257643092057800719L,46051.3999415092985740047109574L,
         -459854.215689014642037955849128L,3.2038910144628771165673418511e6L,
         -1.63159600629029412152839367558e7L,6.28705886134843789288933770703e7L,
         -1.88170788415270893265329732458e8L,4.4628854075147713374039426526e8L,
         -8.51697189501692543040007876698e8L,1.32309878051344428800916830369e9L,
         -1.68747544797148148087387653504e9L,1.77735274675285157416112724604e9L,
         -1.55135562435991186434591782017e9L,1.12345987177168119122825373779e9L,
         -6.74198350156327245555587523995e8L,3.33995951704158160087843043584e8L,
         -1.35648160850208153505036534625e8L,4.46779319816577856556705228286e7L,
         -1.17415681334731764735579417902e7L,2.40319390716489267845190624939e6L,
         -369060.189999853651698913381551L,39996.6708374225632383058837149L,
         -2727.04573891517476624812843511L,87.9692173843604763305847882293L,0L,0L,0L,0L,
         0L},{0L,5.56404391814417356792628705864e-24L,
         -1.19486933281864412845234319438e-14L,2.57755947930153388535254848348e-9L,
         -0.0000128259454234302411119918433211L,0.00807109378518236265942880740185L,
         -1.36361332074427711324656320098L,91.8958001798396057046767715002L,
         -3115.14676364938632248973016752L,61516.1225655751404908864030464L,
         -780892.779191045556369211845372L,6.8279947171256035583835755334e6L,
         -4.3241194583109771764104678482e7L,2.0591018575924337759566083624e8L,
         -7.58618519898637785772210762048e8L,2.2105344296764982135268014692e9L,
         -5.18243019732739760170802334471e9L,9.9062643444050134159630527364e9L,
         -1.55981147067441370064466266143e10L,2.03870425384618943661854562066e10L,
         -2.22395405190227800090590815074e10L,2.03179706309436901446686452033e10L,
         -1.55694761584119873506997233648e10L,1.00026821401329817316148140804e10L,
         -5.37412881000926216420565673104e9L,2.40275417818175525606059217409e9L,
         -8.86979778501030135583915736674e8L,2.67219261707900242529989398577e8L,
         -6.45993730070993835592485510233e7L,1.22242325038800047807274751221e7L,
         -1.7436048693188898661698971182e6L,176235.330877393169268785300119L,
         -11249.0636730250959107735297948L,340.880717364396845781016054389L,0L,0L,0L},
        {0L,8.1824175266826081881268927333e-26L,
         -7.02864313545468103166323354812e-16L,3.41148456723241892388254677067e-10L,
         -3.01826751716675829249919642388e-6L,0.00297070910852734189309320138652L,
         -0.725177922876670708841413638208L,67.0019950224199463921429924177L,
         -3003.06235768926645365481397262L,76410.6242491509177083965554846L,
         -1.22654380941480763386318804619e6L,1.33759590491909302153562474614e7L,
         -1.04587469163411025543628666649e8L,6.10445947153584433760427074844e8L,
         -2.74283638367090440997340983209e9L,9.71674587748060503288076004876e9L,
         -2.76541008486805603593422997477e10L,6.41721181383539961305271837055e10L,
         -1.22842765600939042915497952935e11L,1.95760643765845956112659557169e11L,
         -2.61512562875527464416860745633e11L,2.94341260326461575327951680769e11L,
         -2.80063852279998086455996514355e11L,2.25662281741445254210732768163e11L,
         -1.53992450372740118123290886665e11L,8.88541064262852156846123619083e10L,
         -4.31981388288081569415323979211e10L,1.75938156342177098431106829768e10L,
         -5.95178178503842111152860594459e9L,1.65198999497384614616235509743e9L,
         -3.69716882498492938705775268105e8L,6.50519290881450015150144761122e7L,
         -8.66177902822932385129561794201e6L,820244.226158079910160569880873L,
         -49214.6535694847946096341928524L,1406.13295912813698884669122435L,0L},
        {0L,1.13644687870591780390651287963e-27L,
         -3.90480174208973427161879519732e-17L,4.26436059004270111075957716477e-11L,
         -6.70775865757194766530214189041e-7L,0.00103225079933247109925445413005L,
         -0.363723816880308795196422451522L,45.9916941029039095267224994357L,
         -2718.35013887472646197136716935L,88811.2060060334490621963659206L,
         -1.79502145051696630064656245506e6L,2.42916303331425451841969390302e7L,
         -2.3314422686128816056189348573e8L,1.65687350127903783103004488628e9L,
         -9.0112236975322828121658999545e9L,3.84838735960252474910954533165e10L,
         -1.31702668665544258384166798096e11L,3.67075409981733459980021789267e11L,
         -8.44065834146872803581784031751e11L,1.61785278308565516877105900026e12L,
         -2.60607500172690556494110748834e12L,3.55030427417626640223947576194e12L,
         -4.10978892559509526207003993203e12L,4.05560235399347089199408789018e12L,
         -3.41803044457457258161944681912e12L,2.46142739188776242139526235497e12L,
         -1.5131291979297521779154154821e12L,7.92058354787012196626291928719e11L,
         -3.51539977671286156294264513972e11L,1.31445508404177941892642878657e11L,
         -4.10310324626606737294083268597e10L,1.05571343784255544744725766049e10L,
         -2.1993339675052781923920985907e9L,3.61588535632008419489387407442e8L,
         -4.51441915971753230554623648185e7L,4.02124731873998675956219384098e6L,
         -227617.772758867175069558141942L}
    };
    static const long double varcoeffs[19][37] = {
        {0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L},{0L,0L,0.125L,-0.125L,0.03125L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0L,0L,0.09375L,-0.34375L,0.3984375L,-0.1875L,0.03125L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0L,0L,0.0364583333333333333333333333333L,
         -0.356770833333333333333333333333L,1.02864583333333333333333333333L,
         -1.318359375L,0.8486328125L,-0.26953125L,0.03369140625L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0L,0L,0.009765625L,-0.231770833333333333333333333333L,1.39990234375L,
         -3.65478515625L,5.02286783854166666666666666667L,-3.9306640625L,
         1.7657470703125L,-0.4248046875L,0.04248046875L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{0L,0L,0.00201822916666666666666666666667L,
         -0.111759440104166666666666666667L,1.31245524088541666666666666667L,
         -6.229522705078125L,15.3724212646484375L,-22.22406005859375L,
         19.91118621826171875L,-11.21978759765625L,3.8763427734375L,
         -0.751190185546875L,0.06259918212890625L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L},{0L,0L,0.000341796875L,
         -0.0433844672309027777777777777778L,0.953392537434895833333333333333L,
         -7.76216074625651041666666666667L,31.6900574578179253472222222222L,
         -75.3601430257161458333333333333L,113.191380182902018229166666667L,
         -112.0575714111328125L,74.2905120849609375L,-32.70114898681640625L,
         9.1818523406982421875L,-1.490020751953125L,0.1064300537109375L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0L,0L,0.0000492156498015873015873015873016L,
         -0.0141948760501922123015873015873L,0.571469594561864459325396825397L,
         -7.71698284149169921875L,49.762047672271728515625L,
         -182.9534488677978515625L,423.7363819599151611328125L,
         -655.688915252685546875L,701.00915622711181640625L,
         -525.76217365264892578125L,276.3229248523712158203125L,
         -99.775764942169189453125L,23.5981786251068115234375L,
         -3.2942676544189453125L,0.20589172840118408203125L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L},{0L,0L,6.17617652529761904761904761905e-6L,
         -0.00402954041011749751984126984127L,0.293585609254382905505952380952L,
         -6.4407040618714832124255952381L,63.7300251487701658218625992063L,
         -348.824205875396728515625L,1186.647945225238800048828125L,
         -2694.28946399688720703125L,4269.6773281574249267578125L,
         -4852.6603367328643798828125L,4011.285778820514678955078125L,
         -2417.48377096652984619140625L,1051.809332549571990966796875L,
         -321.900808811187744140625L,65.7670582830905914306640625L,
         -8.055574893951416015625L,0.4475319385528564453125L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L},{0L,0L,6.87587408371913580246913580247e-7L,
         -0.00101234791981059617676642416226L,0.132528479079300343885203097443L,
         -4.66118966575179781232561383929L,69.4383811609100097070926080936L,
         -551.939295116161543225485181052L,2672.76559773870995120396689763L,
         -8561.85488981505235036214192708L,19139.7216060807307561238606771L,
         -30927.9363739416003227233886719L,36947.1820323783904314041137695L,
         -33047.3339365795254707336425781L,22223.1829977594316005706787109L,
         -11184.3174554407596588134765625L,4149.99322288669645786285400391L,
         -1101.9972212612628936767578125L,198.167045228183269500732421875L,
         -21.622891165316104888916015625L,1.08114455826580524444580078125L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{0L,0L,6.88260194485780423280423280423e-8L,
         -0.000228377458174602905488729056437L,0.0534911308441527936824415095899L,
         -2.9902016383232105345953078497L,66.1998807715806564746500112813L,
         -751.155635278727169390077944155L,5067.8221271348343521514266887L,
         -22307.4596401377921066586933439L,68107.9132764380366083175416977L,
         -150358.057820359244942665100098L,246939.991142885293811559677124L,
         -307476.99259200878441333770752L,293597.524055211804807186126709L,
         -216038.005327843129634857177734L,122312.308102918323129415512085L,
         -52814.2493655905127525329589844L,17079.3350669089704751968383789L,
         -4004.97204693965613842010498047L,643.128057417925447225570678711L,
         -63.2684935815632343292236328125L,2.87584061734378337860107421875L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{0L,0L,6.25996897766103495270161936829e-9L,
         -0.0000467797180559173768694744443442L,0.019553737658786314008173622228L,
         -1.72741880394030539762406122117L,56.3220891723128173638273168493L,
         -901.209731678165909316804673937L,8341.91376226235243920532483903L,
         -49520.5811768089021955217633929L,201895.166053194480962933055938L,
         -592650.270112643087105382056463L,1.29493997219509837635245824617e6L,
         -2.15612946263883227948099374771e6L,2.78030081958833913085982203484e6L,
         -2.80521580279043549671769142151e6L,2.22605465388374324538744986057e6L,
         -1.38960495296217966824769973755e6L,678911.501507498323917388916016L,
         -256615.144903259351849555969238L,73527.8725331911118701100349426L,
         -15436.045801985892467200756073L,2239.14511196926468983292579651L,
         -200.544347794493660330772399902L,8.3560144914372358471155166626L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L},{0L,0L,5.21791503084227693602693602694e-10L,
         -8.77792502412334881692266268004e-6L,0.00653800378068048176674462863224L,
         -0.909312714452086503624782283967L,43.3852628514798161400544077013L,
         -970.288775030648741522899170191L,12183.6584683058751433696827771L,
         -96207.4435527823451492521497938L,515234.717838008621345377630658L,
         -1.97228044891148689591015378634e6L,5.6026360624155497602057746715e6L,
         -1.21363933059712262106283257405e7L,2.04465451535879066395517882137e7L,
         -2.7167249253822183818556368351e7L,2.87337442316527192378998734057e7L,
         -2.4316556813545076991431415081e7L,1.64863239976152032031677663326e7L,
         -8.92977636681924050208181142807e6L,3.83533101082151726586744189262e6L,
         -1.2887503214373801893088966608e6L,331455.710904865336488001048565L,
         -62985.5788474131259135901927948L,8331.28191730083926813676953316L,
         -684.825935839035082608461380005L,26.339459070732118561863899231L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L},{0L,0L,4.01427087608050064647286869509e-11L,
         -1.51989064614568082170944423074e-6L,0.00201534203563759843580751027824L,
         -0.440168399592257996155607579934L,30.5926656898878179533683039524L,
         -950.024056279430982558000268358L,16045.7116149736389684910660214L,
         -166796.152429372400715860917612L,1.15895351081197478666177226437e6L,
         -5.70179819116662300584721378982e6L,2.07030969942266226149984976035e7L,
         -5.7197892449642828118360032224e7L,1.22993835411452036501739510439e8L,
         -2.09331595904647331869152064125e8L,2.85446715858638424183178964692e8L,
         -3.14479430894722245284356176853e8L,2.81334758970483623979816911742e8L,
         -2.04757497688241235100576886907e8L,1.21080113748925709954164631199e8L,
         -5.78913990858731841308326693252e7L,2.21765146632996209655175334774e7L,
         -6.7062413118122571358981076628e6L,1.56448530563212536208084202372e6L,
         -271538.645633088435715762898326L,33007.5662774831471324432641268L,
         -2507.08580163022816122975200415L,89.5387786296510057582054287195L,0L,0L,0L,0L,
         0L,0L,0L,0L},{0L,0L,2.86751136974496582011462963844e-12L,
         -2.44323886466255260955007419934e-7L,0.000576429147482515277333699602953L,
         -0.197372282679828950172527772656L,19.9176283748309083760655280016L,
         -854.643463261079095695031150108L,19290.5695822951518477094833981L,
         -261875.028482556511761663813536L,2.33838746995558536395664991405e6L,
         -1.46218025178269136276867357083e7L,6.69900411642232090384065183268e7L,
         -2.3254932080344562348909676075e8L,6.27313935361877476527297403663e8L,
         -1.3404072804628745493755559437e9L,2.30172808141581079705565571203e9L,
         -3.21066385284160685685856151395e9L,3.66570721836012394692230979369e9L,
         -3.44225635366738921463820588542e9L,2.6645938146201618172437974863e9L,
         -1.69973274509648975350273758522e9L,8.90731448309390247633245962788e8L,
         -3.81054067108570516279542061966e8L,1.31706136211956545596990508784e8L,
         -3.62023620895928814888975466602e7L,7.72701676359244420382310636342e6L,
         -1.23418714365893657713968423195e6L,138781.804163323201350976887625L,
         -9796.93364377135367249138653278L,326.564454792378455749712884426L,0L,0L,0L,0L,
         0L,0L},{0L,0L,1.9117325897257989995258513777e-13L,
         -3.66531688326958284438241200111e-8L,0.000153812170520876989468260756875L,
         -0.0824733444080024050434258435858L,12.0559483014173675705332740475L,
         -712.164761501674567127977298079L,21375.2005841031473771119015729L,
         -376590.836998409958518705493289L,4.2892921754515932992462178317e6L,
         -3.37916796408455176669584829559e7L,1.93384420280534853393702346752e8L,
         -8.33797368914608683934644280574e8L,2.78448425459835520496657317207e9L,
         -7.3570526268231073565497808886e9L,1.56336392349168690737384650902e10L,
         -2.70574023224423168618295676424e10L,3.85048125298853813612936392019e10L,
         -4.53677837009358257719236462435e10L,4.44613728263775430378968849254e10L,
         -3.63318282917660680268937767323e10L,2.47631961794556873775135841242e10L,
         -1.40514183277101910066505752184e10L,6.60877684586641384533365339848e9L,
         -2.55740333597376379753995934152e9L,8.05133436440248039697564763628e8L,
         -2.02827301883717225881298418244e8L,3.98954938723564887364858577712e7L,
         -5.90155982949559353656354687701e6L,617352.282067675075083457159053L,
         -40706.4794597892216643231222406L,1272.07748311841317701009757002L,0L,0L,0L,0L},
        {0L,0L,1.19485110083773534953160449192e-14L,
         -5.15470640461631130845144789372e-9L,0.0000384675973690769374844983347111L,
         -0.0322757483157010796200762180768L,6.82291599148941843383502362372L,
         -553.318992521368032772662858423L,21998.5356266570662013567048003L,
         -500524.036015539967483267435385L,7.22854156997041589703963526234e6L,
         -7.12501588995372819687163103351e7L,5.05254653590967881107338440933e8L,
         -2.68102653388528152379217966895e9L,1.09692070876160280073783463589e10L,
         -3.5416567749854548058038433499e10L,9.18891758728025988989209769067e10L,
         -1.9431913972479017004258249699e11L,3.38664587227572184884763618129e11L,
         -4.90590715492137633008972272819e11L,5.94433507263126387564366226002e11L,
         -6.05082167012556221192307532419e11L,5.18751201701960268063519748694e11L,
         -3.74866235683517129189682748347e11L,2.28093278574761002458591541142e11L,
         -1.16505876393678660864929241825e11L,4.96886521553204188812301822509e10L,
         -1.75499656894124538698758186683e10L,5.07243480734316330509825476724e9L,
         -1.17928370463179018330723835106e9L,2.15083529825541683111378432613e8L,
         -2.96275401695758953977133387525e7L,2.89728969402609515015889485312e6L,
         -179220.823975168038089123001555L,5271.20070515200112026832357515L,0L,0L},
        {0L,0L,7.02858951157229904085634224912e-16L,
         -6.82266402870394521173966743667e-10L,9.05324568516602611581851648841e-6L,
         -0.0118803275982029747784388056265L,3.62746673069030167182597691494L,
         -403.017797818349891917012174766L,21160.1663814107738971846863005L,
         -619328.551688685076223744690248L,1.1287053426550236554403643549e7L,
         -1.38414318254876032738420373746e8L,1.20836438137715191759912143483e9L,
         -7.83301126877810718528200728696e9L,3.89368587894832855291423682269e10L,
         -1.52184529769756186821474034389e11L,4.76999386606213882020860274509e11L,
         -1.21783814965408540305146527898e12L,2.5643440444834266295055643541e12L,
         -4.49716090942928417207385937582e12L,6.61903507087651229565527618013e12L,
         -8.22344157688875725594957377984e12L,8.65967795159753370531385105768e12L,
         -7.74918541979653565786669741566e12L,5.89908418379388020421959971451e12L,
         -3.81852279157634977010629417915e12L,2.0974416704907748910603978576e12L,
         -9.7380529155865612460955065055e11L,3.7983641597809560601651545475e11L,
         -1.23374177083033823571310005907e11L,3.29556337768864982178737166318e10L,
         -7.11291597061426411078537412891e9L,1.20927303804641175546206455227e9L,
         -1.55852965701245421249243605644e8L,1.43083520493545930279410960395e7L,
         -833523.539437338004287614268151L,23153.4316510371667857670630042L},
        {0L,0L,3.90478684651002804134825172407e-17L,
         -8.5285083248620035712477467452e-11L,2.01210489516128921631735781253e-6L,
         -0.00412833160389634419793098691857L,1.81909085950230434138509720513L,
         -276.439846425744765439957630532L,19122.5039793556127895646009685L,
         -717723.512699996081530850142126L,1.64425257409286258104244268967e7L,
         -2.49711936099118457731395430313e8L,2.66957151690804650240845906999e9L,
         -2.10128903716883265302986264782e10L,1.2604145416469683822300492132e11L,
         -5.91803114085682136983069779681e11L,2.22173755333567764574359462036e12L,
         -6.78296753697587632328921803954e12L,1.70714019913666498830852419256e13L,
         -3.58090439581775287197763763014e13L,6.31525286759439735769179728778e13L,
         -9.42893006750668120353710080449e13L,1.198153619230555413227268957e14L,
         -1.30084228767228266587056092095e14L,1.20978153029578317789450082898e14L,
         -9.64968177502725506351193383858e13L,6.6014413611261569914032340155e13L,
         -3.86806346878987662129291120489e13L,1.93570709667603717604059714065e13L,
         -8.23550589683979649376973386776e12L,2.95903043904256639264278035378e12L,
         -8.89530383463447420866734716149e11L,2.20856132191121105993735095757e11L,
         -4.44799605083222730868072636312e10L,7.08150528877068989027154521931e9L,
         -8.57470595734067846054524131283e8L,7.41823474298092127239244442105e7L}
    };
    static const long double skewcoeffs[19][37] = {
        {3.L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L},{0.875L,-5.5L,2.75L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.2265625L,-6.65625L,21.421875L,-18.09375L,4.5234375L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.0458984375L,-4.984375L,44.22265625L,-115.55859375L,121.1748046875L,
         -55.37109375L,9.228515625L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L},{0.00748494466145833333333333333333L,
         -2.789306640625L,56.2093505859375L,-316.079345703125L,750.49566650390625L,
         -885.738525390625L,550.0006103515625L,-172.447509765625L,21.555938720703125L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.00101292928059895833333333333333L,-1.25926615397135416666666666667L,
         53.0231572469075520833333333333L,-561.101531982421875L,
         2466.10942840576171875L,-5547.3072509765625L,7063.207489013671875L,
         -5309.669952392578125L,2337.969532012939453125L,-558.0869293212890625L,
         55.80869293212890625L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.000116616960555788070436507936508L,
         -0.47869396209716796875L,40.296621799468994140625L,
         -749.90840244293212890625L,5548.7085173130035400390625L,
         -20891.8238697052001953125L,45407.05202579498291015625L,
         -61001.2893276214599609375L,52338.2821061611175537109375L,
         -28762.23290920257568359375L,9805.152922153472900390625L,
         -1889.70808124542236328125L,157.4756734371185302734375L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.000011654474117137767650462962963L,-0.15753246463795818349041005291L,
         25.8344759042931612206514550265L,-812.70945503711700439453125L,
         9594.8358717024326324462890625L,-56374.74516761302947998046875L,
         190884.837172567844390869140625L,-405249.4096825122833251953125L,
         566837.089881569147109985351563L,-536225.750314235687255859375L,
         345394.98007428646087646484375L,-149376.81729805469512939453125L,
         41532.4535305202007293701171875L,-6711.78125131130218505859375L,
         479.412946522235870361328125L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {1.02800795089000116580377810847e-6L,
         -0.045768326118824973938957093254L,14.3890071961733083876352461558L,
         -746.282584292240559108673580109L,13593.6408359290529338140336294L,
         -119401.455845765769481658935547L,596799.640898387879133224487305L,
         -1.87050639842196553945541381836e6L,3.90484064554001670330762863159e6L,
         -5.63768863639964908361434936523e6L,5.75144713185854628682136535645e6L,
         -4.17921401922353357076644897461e6L,2.150799241858178749680519104e6L,
         -766346.315792597830295562744141L,179873.854549337178468704223633L,
         -25026.9663985446095466613769531L,1564.18539990903809666633605957L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {8.11429452840928678159360532407e-8L,
         -0.0119123151839745265466195565683L,7.10612678936079006504129480433L,
         -597.617709928813079992930094401L,16423.0416148178434620300928752L,
         -209622.247754308208823204040527L,1.49352025658822953701019287109e6L,
         -6.62142289729042258113622665405e6L,1.95704181437716981163248419762e7L,
         -4.03486028647377644665539264679e7L,5.97708936470264697913080453873e7L,
         -6.47708517052058363333344459534e7L,5.17651623663742344360798597336e7L,
         -3.04683917743325494229793548584e7L,1.30437078432276723906397819519e7L,
         -3.95023303397017065435647964478e6L,802125.2536724976380355656147L,
         -97982.7686557697597891092300415L,5443.48714754276443272829055786L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {5.79596589007623181886659212099e-9L,
         -0.00280930474904062788164804852198L,3.15886031188824239911304579841L,
         -425.87811810855161479955154752L,17383.8190858998759404852217625L,
         -316136.250587442798709999474268L,3.1343125416359718189582169529e6L,
         -1.90945088419261503530414922843e7L,7.71713987311825488930769516007e7L,
         -2.17835842066086691367672756314e8L,4.44878698059055453931796364486e8L,
         -6.73263936461257004895014688373e8L,7.66622710599277669216098729521e8L,
         -6.61991914808377202425617724657e8L,4.33874700364649926341371610761e8L,
         -2.14357823427169361792039126158e8L,7.85174971086966893308272119612e7L,
         -2.06721486335191540274536237121e7L,3.6985528217368452533264644444e6L,
         -402647.387348632109933532774448L,20132.3693674316054966766387224L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {3.78032850814193035617019190933e-10L,
         -0.000605795671054649920690627325149L,1.27857717182987904553079888934L,
         -274.143425643780227844780754475L,16437.159144436156588025845676L,
         -419887.581628552958569517137394L,5.69344947613846928700132550494e6L,
         -4.66873490393407008187274359876e7L,2.51793477376356557520618177174e8L,
         -9.4541943312380418551349846114e8L,2.57183416076726822087325022689e9L,
         -5.21392678282314736139596789144e9L,8.03487358480416738357689609984e9L,
         -9.5363838401354313907322648447e9L,8.78225595346152011073900212068e9L,
         -6.28846031953500596763478824869e9L,3.48854890553650872666776194819e9L,
         -1.48403138276638317539618583396e9L,4.74851065523436453329850337468e8L,
         -1.10558972638427478095763945021e8L,1.76780255707925570591214636806e7L,
         -1.73572527698159759347618091851e6L,78896.6034991635269761900417507L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {2.26810373185766408401355062691e-11L,
         -0.000120343686175446240967179918702L,0.475536821286310487069886024574L,
         -161.240445431790778547626450351L,14085.8043169854693449548049654L,
         -500161.588230690085509065413914L,9.15704216638604516003689630972e6L,
         -9.95430376715402992676672511152e7L,7.03699090444964080479622176816e8L,
         -3.44204798947495145510402575642e9L,1.21719957899208831600868214602e10L,
         -3.21228850488519350187384523971e10L,6.47476607412014613600327639428e10L,
         -1.01348302748083245455918176958e11L,1.24609620654242303768030592437e11L,
         -1.21181649150999273173538995252e11L,9.34664775356634623585376147048e10L,
         -5.70852696302923664273442909689e10L,2.74310016235558941596873978597e10L,
         -1.02411623092222058270643856304e10L,2.90844494207502674373719742107e9L,
         -6.06911625287816924812318575277e8L,8.77165852533683674038798017136e7L,
         -7.84300346019537840754765056772e6L,326791.810841474100314485440322L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L},{1.25986748137855055741852183747e-12L,
         -0.0000221619887949193854484565676775L,0.163726163441359238462093458465L,
         -87.4375787931289241766993297918L,11062.5397285279138056172757946L,
         -541727.25965287877216270282589L,1.32618971514921121289121161727e7L,
         -1.88915021340931433800906110644e8L,1.72704308446434637362739782027e9L,
         -1.08341049502310320722695361868e10L,4.89140021005190672082546997636e10L,
         -1.64575268621924181785733798015e11L,4.23485473125307937448130040175e11L,
         -8.49655207411681856172686819819e11L,1.34817931539516586630349379195e12L,
         -1.70888315128898687047014365703e12L,1.74150361794423380942534547522e12L,
         -1.43123564502284014008034763066e12L,9.48274973152596067483929865934e11L,
         -5.04491430150664081370501051538e11L,2.13705890103873786526081662629e11L,
         -7.10665379736629443922415134693e10L,1.81400160323518201998129057984e10L,
         -3.42928045708353924623666841853e9L,4.52164710691244124795774794734e8L,
         -3.71133629975315987443096332754e7L,1.42743703836659995170421666444e6L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L},{6.51692658987584392333714885752e-14L,
         -3.8036291524292862470606489078e-6L,0.052504051511313178759378254588L,
         -44.0373198683661502037631557233L,8033.03681432854757626879589548L,
         -539252.1421709853962394122203L,1.75203388045304746697274376241e7L,
         -3.24128084833245630597781337974e8L,3.79201983405059056879558217298e9L,
         -3.01430823978213007979338568025e10L,1.71370230926644702449376970965e11L,
         -7.23608587713227185459911146012e11L,2.33456171094242547169816789472e12L,
         -5.88021800396337123113096312406e12L,1.17537987401216624049404747241e13L,
         -1.8875351533078282004760169388e13L,2.45701170231792508592048794843e13L,
         -2.60791233415396346568471292575e13L,2.26433404516507833101829600589e13L,
         -1.60901729613407403940299096804e13L,9.3349660857460651886773295237e12L,
         -4.39661681621692042739192193945e12L,1.66462210584273890955895289734e12L,
         -4.98928905474279237357691482835e11L,1.15637500507954779509584308683e11L,
         -1.99803662148571819012329142673e10L,2.42203537004720017829189182557e9L,
         -1.83728860710983003138274671784e8L,6.56174502539225011208123827799e6L,0L,0L,
         0L,0L,0L,0L,0L,0L},{3.15430828290460913090909475748e-15L,
         -6.11212400590773922770255924085e-7L,0.0157634869801633087791831187091L,
         -20.7230803090491614410890966505L,5431.93293020878908173017578579L,
         -497556.519430019607249501957144L,2.13293308578188815814220737739e7L,
         -5.08866272444616782034000463859e8L,7.556020455834870717175470676e9L,
         -7.53907259108300684196844727735e10L,5.33925730325869898174745792292e11L,
         -2.794890649215670019862435724e12L,1.11496923153108250035517452333e13L,
         -3.470416106622715935889997102e13L,8.58260639681232267421558262476e13L,
         -1.71034068383404630891536216402e14L,2.77609421437297601161729324391e14L,
         -3.69933245395597979435222352191e14L,4.06943469625495430683178231825e14L,
         -3.70736737996725940726660347987e14L,2.79997305922781428193501159221e14L,
         -1.7507793261361196081981642753e14L,9.02848002416969993155487633681e13L,
         -3.81320234668405903211285090565e13L,1.30480063013528096128237704341e13L,
         -3.55902604965942008318126215261e12L,7.55330408386027846300081099284e11L,
         -1.20168456274470132995052187586e11L,1.34796535054135450430325151444e10L,
         -9.5047478526025539979011663716e8L,3.16824928420085133263372212387e7L,0L,0L,
         0L,0L,0L,0L},{1.43360662614155980716067288734e-16L,
         -9.23292645167705185622606747958e-8L,0.00445054640583093715573891532878L,
         -9.15768937924438676491661706445L,3440.5476782983572290401147391L,
         -428481.13706629502260290744844L,2.41237927655841816130904710481e7L,
         -7.38051125334619306428796188176e8L,1.38180314428564281011504935407e10L,
         -1.71743637884021580028612354168e11L,1.5021386326397149816229836925e12L,
         -9.65335495969947680326159395543e12L,4.70989261804644299562236353109e13L,
         -1.78937513039019229579641302434e14L,5.39940553907584971012193798167e14L,
         -1.31432810212607973303974140774e15L,2.61263518101146964323182237075e15L,
         -4.28136988460356747890725874071e15L,5.82526458141613401923898219692e15L,
         -6.61429513320538108793870952054e15L,6.28734399561015020637266818257e15L,
         -5.01010155797467660704062957057e15L,3.34488386818692471720875494726e15L,
         -1.8661212750979656623999255648e15L,8.65696743112314570466260279413e14L,
         -3.31317318112524041265441464481e14L,1.03399914483864799300442552225e14L,
         -2.5873441490647371651354817233e13L,5.06383038672465526478931202556e12L,
         -7.46456170729975660916076005847e11L,7.79157357853775362701943147884e10L,
         -5.13230936397209815222501108969e9L,1.60384667624128067257031596553e8L,0L,0L,
         0L,0L},{6.13902999935462341423355402026e-18L,
         -1.31578087047974842564908107839e-8L,0.00118613236812067124481241944836L,
         -3.81657350146864434198303241898L,2051.29833569536162731747683013L,
         -346363.386234985840682599271679L,2.55170308981662674172864004267e7L,
         -9.96608075624626493431411721612e8L,2.34010420689704710067789085052e10L,
         -3.60083267495592573733506359746e11L,3.86240022908063663614704362708e12L,
         -3.02333106114255809552135376756e13L,1.78820725763265586672538328731e14L,
         -8.21093805959540262838763728931e14L,2.9898525664188372667202767071e15L,
         -8.78066420603534506857609116185e15L,2.10802582132763831057349854091e16L,
         -4.18163908868043065665937291238e16L,6.91185549405315745911183729564e16L,
         -9.5812503304135801475298879442e16L,1.11911163890336740889330557046e17L,
         -1.10482394136553038618222572114e17L,9.23318328635071027316901324591e16L,
         -6.53185865337347521496229107446e16L,3.90491469819067864071935875838e16L,
         -1.96573101337017752232111463549e16L,8.28435027207520758727672902298e15L,
         -2.89795862477239140462112415252e15L,8.31198749722804663137260624397e14L,
         -1.92099543153371220507437620145e14L,3.48809621488632800096886223881e13L,
         -4.78987729914568465643930554234e12L,4.67492456637937254240020468877e11L,
         -2.88917082763304189751174700617e10L,8.49756125774424087503455001814e8L,0L,
         0L},{2.48959854301550882962801578089e-19L,
         -1.7745942183127210987335303503e-9L,0.000299405373562301408355757403694L,
         -1.50561884893099522861572338385L,1156.00995409874525573804543918L,
         -264064.586990266836913820615205L,2.53814264307375445388015980414e7L,
         -1.26087919204649232262195308017e9L,3.69692028801362787460662683387e10L,
         -7.00720905404805734661343442164e11L,9.16485665885131304846535249941e12L,
         -8.68168207920699247065436676536e13L,6.18011580250398292766502356113e14L,
         -3.40222918850083182569762469496e15L,1.48167155919250813603733920028e16L,
         -5.19796662528819580244646477066e16L,1.49054583263826367296397490359e17L,
         -3.53509671381588992882455505843e17L,7.00026574310203518275281691947e17L,
         -1.16615823229243996921159240169e18L,1.64390188090700172062950937993e18L,
         -1.96955111858253354762721277218e18L,2.01154827367176435607791764841e18L,
         -1.7542436246127813793562047882e18L,1.30675758756765064838898691773e18L,
         -8.30602588736836403863107025372e17L,4.49336272799187098848766163155e17L,
         -2.05994507449074535620192469201e17L,7.95154797338133564735886737207e16L,
         -2.56091870936430568869659061896e16L,6.79449351940530999862331869244e15L,
         -1.4587501408373372812959612296e15L,2.47022700635653973589782500741e14L,
         -3.1747874161394451409880384992e13L,2.90956123256576972006868141148e12L,
         -1.69354305241617779174255755458e11L,4.70428625671160497706265987384e9L}
    };
    static const long double kurtcoeffs[19][37] = {
        {0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L},{8.L,-18.L,9.L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{5.75L,-50.5L,106.25L,
         -81.L,20.25L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {2.95833333333333333333333333333L,-73.8125L,376.46875L,-752.4375L,
         704.203125L,-309.65625L,51.609375L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{1.21354166666666666666666666667L,
         -74.7552083333333333333333333333L,780.893229166666666666666666667L,
         -3086.921875L,6003.01953125L,-6361.6171875L,3745.97265625L,-1151.015625L,
         143.876953125L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.418229166666666666666666666667L,-58.8326171875L,1158.27275390625L,
         -8073.3623046875L,27443.476806640625L,-52534.67626953125L,
         60578.041748046875L,-42965.474609375L,18343.142822265625L,-4324.15283203125L,
         432.415283203125L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.124696180555555555555555555556L,-38.2464789496527777777777777778L,
         1350.85271538628472222222222222L,-15599.7976725260416666666666667L,
         85262.1811482747395833333333333L,-262234.5843505859375L,
         497230.719462076822916666666667L,-609836.845703125L,493163.4617919921875L,
         -261295.7391357421875L,87288.46746826171875L,-16679.6982421875L,
         1389.974853515625L,0L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.0327721974206349206349206349206L,-21.3192432221912202380952380952L,
         1308.22734825497581845238095238L,-24081.1082499186197916666666667L,
         201520.447060648600260416666667L,-934574.599647013346354166666667L,
         2.67721614160105387369791666667e6L,-5.04639351523844401041666666667e6L,
         6.48943063618214925130208333333e6L,-5.790530973907470703125e6L,
         3.5857510322265625e6L,-1.5122900873565673828125e6L,
         414474.823810577392578125L,-66565.53458404541015625L,
         4754.681041717529296875L,0L,0L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.00769488622271825396825396825397L,-10.4457041180323040674603174603L,
         1089.52302622719416542658730159L,-31149.8770693097795758928571429L,
         386642.142534691946847098214286L,-2.599036452996063232421875e6L,
         1.07015771501453399658203125e7L,-2.9075554627471923828125e7L,
         5.46045338601531982421875e7L,-7.2918917646640777587890625e7L,
         7.02747185595874786376953125e7L,-4.90277772784271240234375e7L,
         2.45314121459064483642578125e7L,-8.5815332711200714111328125e6L,
         1.99270953053188323974609375e6L,-275948.57366180419921875L,
         17246.785853862762451171875L,0L,0L,0L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.00163108112530313051146384479718L,-4.57827206843744510065310846561L,
         799.346247725133542661313657407L,-34855.0145927497318812779017857L,
         627748.103215935684385753813244L,-5.95548077365363401079934740823e6L,
         3.41180491829898745294601198227e7L,-1.28357494991866800520155164931e8L,
         3.34787843225419607427385118273e8L,-6.27345246591134428977966308594e8L,
         8.64056468872793058554331461589e8L,-8.86156730595170736312866210937e8L,
         6.79703121283103724320729573568e8L,-3.88240303523262381553649902344e8L,
         1.62729562511351883411407470703e8L,-4.859420606284618377685546875e7L,
         9.78544525733667612075805664063e6L,-1.19087777266037464141845703125e6L,
         66159.876258909702301025390625L,0L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.000314795546220513668430335097002L,-1.81843643205716706667837852734L,
         525.727404226523017546907724317L,-34503.4124356345398716194919808L,
         887496.406858166213388796205874L,-1.16596083639457533283839150081e7L,
         9.08487380291426418477265292375e7L,-4.60541571280199312028430757068e8L,
         1.61410879596226243064517066592e9L,-4.07597660620791527628898620605e9L,
         7.62835433718905102908611297607e9L,-1.07822273836602095286051432292e10L,
         1.16417617846806479950745900472e10L,-9.64934935189494668443997701009e9L,
         6.13062983830293955902258555094e9L,-2.95981535383948349952697753906e9L,
         1.06641908221254836022853851318e9L,-2.7766579567427971959114074707e8L,
         4.93478817640941888093948364258e7L,-5.35610593052119016647338867187e6L,
         267805.296526059508323669433594L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},
        {0.0000557157667244744067660734327401L,
         -0.661090784789192498791040050806L,314.064758510646580961703566074L,
         -30721.9876662795490058010848111L,1.11539835247405243495469370847e6L,
         -2.00101040179989791018937630628e7L,2.08569952211833341263195194265e8L,
         -1.39704619766360606795580929549e9L,6.42869982892890571801474800816e9L,
         -2.12808918985220508507319859096e10L,5.23497607050106681336604413532e10L,
         -9.79195193704881459958851337433e10L,1.41491226964514212600079675515e11L,
         -1.59542598521807158685103058815e11L,1.41092397297485010681860148907e11L,
         -9.78778777323389377593994140625e10L,5.2991038859992506515234708786e10L,
         -2.21337208794477364383637905121e10L,6.98906560754662665911018848419e9L,
         -1.61265562332230789586901664734e9L,2.56451379855482843704521656036e8L,
         -2.51169426699096355587244033813e7L,1.14167921226861979812383651733e6L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{9.09911924256881396638341082786e-6L,
         -0.221733689333662772177375686219L,172.162016985662330092703867338L,
         -24918.517368028283809667522754L,1.26564251830589181723702368645e6L,
         -3.06777275482734500827691542408e7L,4.22441842219308080214389470099e8L,
         -3.68459214817739458732977115288e9L,2.18848560226703483235307786831e10L,
         -9.30928189672431507531050888319e10L,2.94023283138317095571373485857e11L,
         -7.0783858508439473481374600577e11L,1.32388020912549829549752323637e12L,
         -1.94982461868550944269300541944e12L,2.28177535894638972610872135394e12L,
         -2.13226621628328038646740218004e12L,1.5929436539163766181471835201e12L,
         -9.4865172397805568989273160696e11L,4.46993878309076123063607762257e11L,
         -1.64416455051814562057144939899e11L,4.61887191662341339222621172667e10L,
         -9.56661895509222135460004210472e9L,1.37635398730945628439076244831e9L,
         -1.22805348413334882818162441254e8L,5.1168895172222867840901017189e6L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L,0L,0L},{1.37854670626310980043618932508e-6L,
         -0.0690582871400490537335609446103L,87.3045406363051809922596370916L,
         -18594.5476190105814635168288559L,1.31223594567111300266772199226e6L,
         -4.26243227797489459825367128474e7L,7.67990223235081051594211106365e8L,
         -8.62522742391042970270320865107e9L,6.52727082479078912054180644775e10L,
         -3.51480198692716452234865735388e11L,1.40079555935712018490149014353e12L,
         -4.25371259385992278699006160928e12L,1.00573066049423355988948325696e13L,
         -1.88147945763151834041071822867e13L,2.81772216821918849398962959337e13L,
         -3.40538664756233057755537641545e13L,3.33724566908996611840549875827e13L,
         -2.65653281084193850753262871876e13L,1.71527486610337520984234210725e13L,
         -8.93940080995225466996045239891e12L,3.72610690967099202078281571933e12L,
         -1.22388079064069610029869363643e12L,3.09572884758788196777444682084e11L,
         -5.81562913319544964703964069486e10L,7.63847664468249241205921862274e9L,
         -6.25835740702126607095124199986e8L,2.4070605411620254119043238461e7L,0L,0L,
         0L,0L,0L,0L,0L,0L,0L,0L},{1.94668400663808979316171776489e-7L,
         -0.0200794874637445723200741688551L,41.2268593558527859052060148347L,
         -12867.279202521007365136896382L,1.25500596394950898059417775107e6L,
         -5.42795849794215626826337030217e7L,1.26998957216505436446046039078e9L,
         -1.82048505562851921564748980187e10L,1.73763301137229756002955333773e11L,
         -1.17080960131321568936195306753e12L,5.8103537790508249882642643772e12L,
         -2.19192672185884961316456505166e13L,6.43766273797313899447677778813e13L,
         -1.49899754233137506266876504351e14L,2.80557892006263668191033345663e14L,
         -4.26413656187431495294492871922e14L,5.30110538929863735307428851575e14L,
         -5.41520105107795121355113922618e14L,4.55484882172485727670770870949e14L,
         -3.15316724932445254025478890981e14L,1.79081065139922303012027205114e14L,
         -8.29113796549872476606538330088e13L,3.0969328773727590360780381161e13L,
         -9.18602955552618326135579263791e12L,2.11268190168224138865106215235e12L,
         -3.63081856931809203246863035019e11L,4.38666438257563449163244513329e10L,
         -3.32243966896519557511783204973e9L,1.18658559605899841968494001776e8L,0L,0L,
         0L,0L,0L,0L,0L,0L},{2.57302495209279186162039998019e-8L,
         -0.00547555756618909382911170211318L,18.2279530857737818376729038077L,
         -8310.70487163579264500721640719L,1.11565273308130311326437748236e6L,
         -6.3922957027009504319866807242e7L,1.93048418419185936587340222612e9L,
         -3.50734847425555781759200396581e10L,4.18852620229211528332812100802e11L,
         -3.4991301295306893230018922968e12L,2.13987297280018708722077326074e13L,
         -9.91054973583539480834065224328e13L,3.56729642832733641332624810189e14L,
         -1.01810106742848836230234355459e15L,2.3397783535270273593296444652e15L,
         -4.38197307036943079770105694308e15L,6.74826640371632214525469156458e15L,
         -8.60166394282920634062448820689e15L,9.11424434142642195014848505168e15L,
         -8.04609895780190963628098890654e15L,5.91913599549099504996506381071e15L,
         -3.6213636175620332889972704795e15L,1.83437288676157293876779734625e15L,
         -7.63611478722774855216371051938e14L,2.58300907521124540540998219361e14L,
         -6.98296429749599755270155910694e13L,1.47217782343440935323672902086e13L,
         -2.33128235024003260727795077401e12L,2.60750528715975808767639364305e11L,
         -1.83610804943619090081870126596e10L,6.12036016478730300272900421987e8L,0L,
         0L,0L,0L,0L,0L},{3.19520876656535576662193932366e-9L,
         -0.00140590331614618414089550814137L,7.580617941244477462981814022L,
         -5036.99033345287919163864721923L,927643.090396631102643098017761L,
         -7.01265815344914827320648842829e7L,2.72037335769402087150977781205e9L,
         -6.22881586662858637270040949333e10L,9.24639270932976650987041423139e11L,
         -9.5066510866254138042548275554e12L,7.10449703811928582016208811749e13L,
         -4.00158930448491075484492235332e14L,1.74660860704150912598802569339e15L,
         -6.03689085051499681274795213997e15L,1.68056462993092205822744841668e16L,
         -3.81871832275099986236417971491e16L,7.15703844283188503686178293754e16L,
         -1.11528633047911631682739616627e17L,1.45364473715521009807443054162e17L,
         -1.59119139599891797453239105614e17L,1.46620617021737128404985924612e17L,
         -1.13801179254329241968257989553e17L,7.43144171987483705992995374703e16L,
         -4.07020701805561594294838004162e16L,1.85962148316866421686998456547e16L,
         -7.02932569386071553774076694079e15L,2.17211338607626648250652409096e15L,
         -5.39341349048801616988944274908e14L,1.04949150021051838249210285994e14L,
         -1.54077759143810411690806955676e13L,1.60419382109009379170472055876e12L,
         -1.05542060381683137361486842565e11L,3.29818938692759804254646383015e9L,0L,
         0L,0L,0L},{3.7405179940969095608341172018e-10L,
         -0.000341058893042362996494000145569L,2.97703135970679976754435828799L,
         -2877.71973536126285368504904188L,725210.642929437617995285256069L,
         -7.20979681566450054708034266699e7L,3.57842846958443347284611585595e9L,
         -1.02783616968380141718341456461e11L,1.88652552096407418153774327938e12L,
         -2.3727206424626377053438793832e13L,2.15212841806151405948899269215e14L,
         -1.46296124320732533111570987895e15L,7.67706116964737408476046260856e15L,
         -3.18280540016805768984586233968e16L,1.06177992184400019739674159413e17L,
         -2.89204380098578075297297250455e17L,6.50687546953780500190656035711e17L,
         -1.22050625106445898162642379805e18L,1.92233362870043491954247937623e18L,
         -2.55620413365125949663970716465e18L,2.88077773728995587368979535978e18L,
         -2.75802362924071027778189282955e18L,2.24521807970591895640190932423e18L,
         -1.55326596483869373800889259248e18L,9.11214090471365983222247833556e17L,
         -4.51497634373692001950570800978e17L,1.87794367202946864785295279745e17L,
         -6.49899284234254228541406342564e16L,1.84802980843435165200726940591e16L,
         -4.24227384638473419697196429575e15L,7.6640276862779842961916117034e14L,
         -1.04866732843826181295620848118e14L,1.0212003732587762474983913652e13L,
         -6.30447121019835846317796558935e11L,1.85425623829363484211116634981e10L,0L,
         0L},{4.14069524749690535878321599767e-11L,
         -0.0000784097605054450041268271040522L,1.10776271042037954787887328213L,
         -1555.75842819568893906946903638L,535410.177868359456197928309736L,
         -6.98165862062761167069055586908e7L,4.41927218943611708359891045292e9L,
         -1.58628773572912956412108894613e11L,3.58411365247430994064930554336e12L,
         -5.4868270719551690047045937835e13L,6.00636818795669115007311491934e14L,
         -4.89671277279540105676113596778e15L,3.06770839155571747610354004854e16L,
         -1.51366306615637133169140721151e17L,5.99865288910615739830865021711e17L,
         -1.9396104433451221362010581919e18L,5.18225495099853483657761859285e18L,
         -1.15587038909565523452897708761e19L,2.16993927783153418650052999907e19L,
         -3.45102494883600427047417630987e19L,4.67273241844774108111880641633e19L,
         -5.40612924397892169631318245699e19L,5.35690411641217209960986825107e19L,
         -4.55131679922781128323500238409e19L,3.31510517966616632091904037786e19L,
         -2.06709866449997336776736783473e19L,1.10016491369294470159839211221e19L,
         -4.97479229688973768822164152013e18L,1.89844195317933640294692861747e18L,
         -6.05695420535318392061742085408e17L,1.59485318419150793117299038623e17L,
         -3.40375684543189911456294965717e16L,5.73803797853060886888649091341e15L,
         -7.35121982385931886495025168012e14L,6.72359118069673864458390785055e13L,
         -3.90982945979233425041219370044e12L,1.08606373883120395844783158346e11L}
    };
    long double ocoeffs[19], derivcoeffs[19];
    size_t i;
    const long double twof0 = 2.0L*f0;

    if (maxdeg < 1)
        maxdeg = 1;
    if (maxdeg > 18)
        maxdeg = 18;

    /* Check the applicability of the series expansion */
    assert(initialized);
    *ifail = 1;
    if (o > 0.25 || o < 0.0)
    {
        *mean = 0.0;
        *var = 0.0;
        *skew = 0.0;
        *dskewdo = 0.0;
        *dskewdf0 = 0.0;
        *kurt = 0.0;
        *dkurtdo = 0.0;
        *dkurtdf0 = 0.0;
        return;
    }

    /* Calculate the mean */
    for (i=0; i<=maxdeg; ++i)
        ocoeffs[i] = poly_sum_ld(&meancoeffs[i][0], maxdeg*2, twof0);
    *mean = (double)poly_sum_ld(ocoeffs, maxdeg, o);

    /* Calculate the variance */
    for (i=0; i<=maxdeg; ++i)
        ocoeffs[i] = poly_sum_ld(&varcoeffs[i][0], maxdeg*2, twof0);
    *var = (double)poly_sum_ld(ocoeffs, maxdeg, o);

    /* Skewness calculation */
    if (o > 0.0)
    {
        const double a = 1.0/sqrt(o);
        long double tmp, ldskew, lddskewdo, lddskewdf0, fac;

        for (i=0; i<=maxdeg; ++i)
            poly_and_deriv(&skewcoeffs[i][0], maxdeg*2, twof0,
                           ocoeffs+i, derivcoeffs+i);
        poly_and_deriv(ocoeffs, maxdeg, o, &tmp, &lddskewdo);
        lddskewdf0 = poly_sum_ld(derivcoeffs, maxdeg, o);

        fac = (1.0L - twof0)/a/SQR2;
        ldskew = fac*tmp;
        lddskewdo = fac*(lddskewdo + tmp/o/2.0L);
        lddskewdf0 = 2.0L*(lddskewdf0*fac - tmp/a/SQR2);

        *skew = (double)ldskew;
        *dskewdo = (double)lddskewdo;
        *dskewdf0 = (double)lddskewdf0;
    }
    else
    {
        *skew = 0.0;
        *dskewdo = 0.0;
        *dskewdf0 = 0.0;
    }

    /* Kurtosis calculation */
    {
        long double ldkurt, lddkurtdo, lddkurtdf0;

        for (i=0; i<=maxdeg; ++i)
            poly_and_deriv(&kurtcoeffs[i][0], maxdeg*2, twof0,
                           ocoeffs+i, derivcoeffs+i);
        poly_and_deriv(ocoeffs, maxdeg, o, &ldkurt, &lddkurtdo);
        lddkurtdf0 = poly_sum_ld(derivcoeffs, maxdeg, o);
        lddkurtdf0 *= 2.0L;
        ldkurt += 3.0L;

        *kurt = (double)ldkurt;
        *dkurtdo = (double)lddkurtdo;
        *dkurtdf0 = (double)lddkurtdf0;
    }

    *ifail = 0;
}

static void big_a_shape(const double a, const double signedb,
                        size_t maxdeg, int *ifail,
                        double *mean, double *var,
                        double *skew, double *kurt,
                        double *dskewdo, double *dskewdf0,
                        double *dkurtdo, double *dkurtdf0)
{
    const double b = fabs(signedb);
    const double o = 1.0/a/a;
    const long double f0 = b/a > max_exponent ? 0.L :
        1.0L/(1.0L + expl((long double)(b/a)));
    big_a_shape_natural_pars(o, f0, maxdeg, ifail, mean, var, skew,
                             kurt, dskewdo, dskewdf0, dkurtdo, dkurtdf0);
}

static void zero_skew_kurtosis(const double o_in, size_t maxdeg,
                               int *ifail, double *kurt, double *dkurtdo)
{
    /* Argument o_in should be set to 1/a^2.
     * The following series converge well for a >= 2.0.
     * a = 2.0 corresponds to kurtosis value of about 2.78851.
     */
    static long double ocoeffs[19] = {
        0.L,-1.L,0.75L,-0.666666666666666666666666666667L,0.666015625L,
        -0.730989583333333333333333333333L,0.870552571614583333333333333333L,
        -1.1161835201202876984126984127L,1.53234282448178245907738095238L,
        -2.24311088195343286692570546737L,3.48941642193251816683976107804L,
        -5.75180598073301358121278979026L,10.0203665926068479484026298627L,
        -18.4062217128730021663377100762L,35.5707483799407807007176541246L,
        -72.1728020413790148936248797921L,153.448812509861417299856889262L,
        -341.246672090303143837451790974L,792.390977024097729233967929523L
    };
    long double ldkurt, lddkurtdo;
    const long double o = o_in;

    if (maxdeg < 1)
        maxdeg = 1;
    if (maxdeg > 18)
        maxdeg = 18;
    poly_and_deriv(ocoeffs, maxdeg, o, &ldkurt, &lddkurtdo);
    ldkurt += 3.0L;

    *ifail = 0;
    *kurt = ldkurt;
    *dkurtdo = lddkurtdo;
}

void sbfitmod_(const int *dumm, const double *mean,
               const double *sigma, const double *skew,
               const double *kurt, double *gamma, double *delta,
               double *xlam, double *xi, int *fault)
{
    /* Try to see whether we need to invoke some
     * special case before calling the general routine
     */
    const double max_high_a_skew = 0.730697;
    const int maxiter = 500;
    const double bigkurt = 2.78851;
    double dtmp, w, lognormkurt;
    const double b1 = *skew * *skew;
    const double rb1 = fabs(*skew);
    int niter, ifail;

    /* Initialize static data */
    if (!initialized)
        john_init();

    /* Check that we are in the S_b range */
    dtmp = pow(((2.0+b1+sqrt(b1*(4.0+b1)))/2.0), 1.0/3.0);
    w = dtmp + 1.0/dtmp - 1.0;
    lognormkurt = w*w*(w*(w+2.0)+3.0)-3.0;

    if (*kurt >= lognormkurt || *kurt <= b1 + 1.0)
    {
        /* Wrong function */
        printf("ERROR in sbfitmod_: skew, kurtosis not in S_b range\n");
        *fault = 1;
        return;
    }

    /* Check for 0 skewness and high kurtosis */
    if (*skew == 0.0 && *kurt >= bigkurt)
    {
        const double tol = 1.0e-12;
        const int maxdeg_0skew = 18;
        double asq, diff, trykurt, dkurtdo, hmu2;

        /* Initial guess for the value of o = 1/a^2 = 2/delta^2 */
        dtmp = pow(55890.0 + sqrt(35478972.0+pow((55890.0-*kurt*20736.0),2))
                   - *kurt*20736.0, 1.0/3.0);
        asq = 0.375 - 10.866819/dtmp + dtmp/30.2381052;

        /* Iterations to get the correct kurtosis */
        for (niter=0; niter<maxiter; ++niter)
        {
            zero_skew_kurtosis(asq, maxdeg_0skew,
                               &ifail, &trykurt, &dkurtdo);
            assert(ifail == 0);
            diff = trykurt - *kurt;
            if (fabs(diff) <= tol)
                break;
            asq -= diff/dkurtdo;
        }
        assert(niter < maxiter);

        /* The upper limit in the following assertion
         * should be synchronized with the "bigkurt" value
         */
        assert(asq > 0.0 && asq < 0.26);

        /* Calculate the distribution parameters */
        *delta = sqrt(2.0/asq);
        *gamma = 0.0;
        {
            static const long double varcoeffs[19] = {
                0.L, 0.03125L, -0.0078125L,
                0.00276692708333333333333333333333L, 
                -0.00126139322916666666666666666667L, 
                0.000702921549479166666666666666667L, 
                -0.000462934705946180555555555555556L, 
                0.000351788127233111669146825396825L, 
                -0.000302970504003857809399801587302L, 
                0.000291624635037501232544160604056L, 
                -0.000310251408534180340102530450838L, 
                0.000361502959292694981212472235218L, 
                -0.000457848846570355307491399942284L, 
                0.000626262125360912249491673548415L, 
                -0.000920077487273003610733353602142L, 
                0.00144496177082410849596712187139L,
                -0.00241568640947414568470574648008L,
                0.00428330361054101555720953273142L,
                -0.00802880374681102119742268410718L
            };
            hmu2 = poly_sum_ld(varcoeffs, maxdeg_0skew,
                               (long double)asq);
        }
        *xlam = *sigma/sqrt(hmu2);
        *xi = *mean - *xlam * 0.5;
        *fault = 0;
        return;
    }

    /* Check for small skewness and high kurtosis
     * (this is the "large a" mode)
     */
    if (rb1 <= max_high_a_skew)
    {
        static const double high_a_coeffs[4] = {
            2.89695167542, 0.0207685492933,
            1.78114676476, 0.258008509874
        };
        const double bigakurt = poly_sum(high_a_coeffs,
            sizeof(high_a_coeffs)/sizeof(high_a_coeffs[0])-1, rb1);
        if (*kurt >= bigakurt)
        {
            /* Perform pure "large a" iterations */
            const int large_a_maxdeg = 18;
            double o, f0, trymean, tryvar, tryskew, trykurt;
            double dskewdo, dskewdf0, dkurtdo, dkurtdf0;
            int converged = 0;

            /* The iterations may be sufficiently slow
             * so that it makes sense to save the answer
             */
            static double old_skew = 0.0, old_kurt = -1.0,
                old_o = 100.0, old_f0 = -1.0;

            if (*skew == old_skew && *kurt == old_kurt)
            {
                o = old_o;
                f0 = old_f0;
                big_a_shape_natural_pars(o, (long double)f0, large_a_maxdeg,
                                         &ifail, &trymean, &tryvar, &tryskew, &trykurt,
                                         &dskewdo, &dskewdf0, &dkurtdo, &dkurtdf0);
                assert(ifail == 0);
                converged = 1;
            }
            else
            {
                /* First approximation for o and f0 */
                const double tol = 1.0e-10;
                const double maxupratio = 2.0, maxdownratio = 3.0;
                const double excess = *kurt - 3.0;
                double tsq, diffskew, diffkurt, new_o, new_f0;

                if (excess != 0.0)
                {
                    double r = b1/excess;
                    tsq = 2.0*r/(2.0*r - 1.0)/9.0;
                }
                else
                    tsq = 1.0/9.0;
                assert(tsq > 0.0);
                if (tsq < 1.0)
                    f0 = (1 - sqrt(tsq))/2.0;
                else
                {
                    tsq = 1.0;
                    f0 = 5.0e-5;
                }
                o = 2.0*b1/9.0/tsq;

                /* Improve o and f0 by iterations */
                for (niter=0; niter<maxiter; ++niter)
                {
                    big_a_shape_natural_pars(o, (long double)f0, large_a_maxdeg,
                                             &ifail, &trymean, &tryvar, &tryskew, &trykurt,
                                             &dskewdo, &dskewdf0, &dkurtdo, &dkurtdf0);
                    assert(ifail == 0);

                    diffskew = tryskew - rb1;
                    diffkurt = trykurt - *kurt;
                    if (fabs(diffskew) < tol && fabs(diffkurt) < tol)
                    {
                        converged = 1;
                        old_skew = *skew;
                        old_kurt = *kurt;
                        old_o = o;
                        old_f0 = f0;
                        break;
                    }
                    dtmp = dkurtdo*dskewdf0 - dkurtdf0*dskewdo;
                    if (dtmp == 0.0)
                    {
                        /* Null determinant. Can't iterate anymore.
                         * Hope that this can be solved by the generic
                         * routine which will use different variables.
                         */
                        break;
                    }

                    new_o  = o + (dkurtdf0*diffskew - dskewdf0*diffkurt)/dtmp;
                    if (new_o > 0.25)
                        o = 0.25;
                    else if (new_o > o*maxupratio)
                        o *= maxupratio;
                    else if (new_o < o/maxdownratio)
                        o /= maxdownratio;
                    else
                        o = new_o;
                
                    new_f0 = f0 + (dskewdo*diffkurt - dkurtdo*diffskew)/dtmp;
                    if (new_f0 > 0.5)
                        f0 = 0.5;
                    else if (new_f0 > f0*maxupratio)
                        f0 *= maxupratio;
                    else if (new_f0 < f0/maxdownratio)
                        f0 /= maxdownratio;
                    else
                        f0 = new_f0;
                }
            }

            if (converged)
            {
                assert(o > 0.0);
                *delta = SQR2/sqrt(o);

                assert(f0 >= 0.0 && f0 <= 0.5);
                if (f0 == 0.0)
                    *gamma = *delta * 200.0;
                else if (f0 == 0.5)
                    *gamma = 0.0;
                else
                {
                    *gamma = *delta * log(1.0/f0 - 1.0);
                    if (*gamma < 0.0)
                        *gamma = 0.0;
                }
                assert(tryvar > 0.0);
                *xlam = *sigma/sqrt(tryvar);
                if (*skew > 0.0)
                    *xi = *mean - *xlam * trymean;
                else
                {
                    *gamma *= -1.0;
                    *xi = *mean - *xlam * (1.0 - trymean);
                }
                *fault = 0;
                return;
            }
            else
            {
                printf("WARNING in sbfitmod_: no large a convergence "
                       "for skew = %g, kurt = %g\n", *skew, *kurt);
            }
        }
    }

    /* Try the "large gamma" mode */
    sbfitlargegamma_(mean, sigma, skew, kurt,
                     gamma, delta, xlam, xi, fault);
    if (*fault == 0)
        return;

    /* Finally, call the generic fitter */
    sbfit_(mean, sigma, skew, kurt,
           gamma, delta, xlam, xi, fault);
    if (*fault)
        printf("ERROR in sbfitmod_: default algorithm failed "
               "for skew = %g, kurt = %g\n", *skew, *kurt);
}

void johnintf_(const double *gamma, const double *delta,
	       double *pmean, double *pvar, double *pskew, double *pkurt,
               double *pdskewddelta, double *pdskewdgamma,
               double *pdkurtddelta, double *pdkurtdgamma)
{
    double weight[2];
    double mean[2] = {0.0, 0.0}, var[2] = {0.0, 0.0};
    double skew[2] = {0.0, 0.0}, kurt[2] = {0.0, 0.0};
    double dskewddelta[2] = {0.0, 0.0}, dskewdgamma[2] = {0.0, 0.0};
    double dkurtddelta[2] = {0.0, 0.0}, dkurtdgamma[2] = {0.0, 0.0};
    const double a = *delta/SQR2;
    const double b = *gamma/SQR2;

    /* Initialize static data */
    if (!initialized)
        john_init();

    /* Interpolate weight smoothly between a == 2 and a == 3 */
    if (a <= 2.0)
        weight[1] = 0.0;
    else if (a < 3)
    {
        const double log2 = 0.69314718055994530942;
        const double log3 = 1.0986122886681096914;
        const double loga = log(a);
        double logfrac = (loga - log2)/(log3 - log2);
        if (logfrac < 0.0) logfrac = 0.0;
        if (logfrac > 1.0) logfrac = 1.0;
        weight[1] = logfrac*logfrac*(3.0 - 2.0*logfrac);
    }
    else
        weight[1] = 1.0;
    weight[0] = 1.0 - weight[1];

    if (weight[0] > 0.0)
    {
        /* Classic method */
        double factor, logfactor;

        johnint(a, b, 0, mean, var, skew, kurt, &logfactor,
                dskewddelta, dskewdgamma, dkurtddelta, dkurtdgamma);
        factor = exp(logfactor);
        *mean *= factor;
        *var  *= (factor*factor);
        *dskewdgamma *= factor;
        *dkurtdgamma *= factor;
    }

    if (weight[1] > 0.0)
    {
        /* Large a method */
        int ifail;
        const size_t big_a_deg = 18;
        const double aderfactor  = -2.0/a/a/a;
        const double absb = fabs(b);
        double bderfactor = 0.0, crossfactor = 0.0;

        big_a_shape(a, b, big_a_deg, &ifail, mean+1, var+1,
                    skew+1, kurt+1, dskewddelta+1,
                    dskewdgamma+1, dkurtddelta+1, dkurtdgamma+1);
        assert(ifail == 0);

        /* Convert from derivatives over o = 1/a^2 and
         * f0 = 1/(1 + exp(absb/a)) into derivatives over a and absb.
         */
        if (absb/a < max_exponent)
        {
            const double eba = exp(absb/a);
            bderfactor  = -eba/a/(1.0 + eba)/(1.0 + eba);
            crossfactor = -absb/a*bderfactor;
        }
        dskewddelta[1] = dskewddelta[1]*aderfactor + dskewdgamma[1]*crossfactor;
        dskewdgamma[1] *= bderfactor;
        dkurtddelta[1] = dkurtddelta[1]*aderfactor + dkurtdgamma[1]*crossfactor;
        dkurtdgamma[1] *= bderfactor;

        if (b == 0.0)
        {
            skew[1] = 0.0;
            dkurtdgamma[1] = 0.0;
            dskewddelta[1] = 0.0;
        }
        else if (b < 0.0)
        {
            skew[1] *= -1.0;
            dkurtdgamma[1] *= -1.0;
            dskewddelta[1] *= -1.0;
        }
    }

    /* Create weighted result */
    *pmean = mean[0]*weight[0] + mean[1]*weight[1];
    *pvar  = var[0]*weight[0]  + var[1]*weight[1];
    *pskew = skew[0]*weight[0] + skew[1]*weight[1];
    *pkurt = kurt[0]*weight[0] + kurt[1]*weight[1];
    *pdskewddelta = (dskewddelta[0]*weight[0] + dskewddelta[1]*weight[1])/SQR2;
    *pdskewdgamma = (dskewdgamma[0]*weight[0] + dskewdgamma[1]*weight[1])/SQR2;
    *pdkurtddelta = (dkurtddelta[0]*weight[0] + dkurtddelta[1]*weight[1])/SQR2;
    *pdkurtdgamma = (dkurtdgamma[0]*weight[0] + dkurtdgamma[1]*weight[1])/SQR2;
}

void johnint(double a, double b, double *p, double *mean, double *var,
	     double *skew, double *kurt, double *logfactor, double *dskewda,
             double *dskewdb, double *dkurtda, double *dkurtdb)
{
    /* This function calculates the integrals
     * 1/sqrt(Pi)/(1+exp(-(x-|b|)/|a|))^n * exp(-x^2)
     * from -Infinity to Infinity for n = 1 ... 6.
     */
    int prec, npos, nneg, j, positive_variance;
    int b_is_negative = 0, large_ratio = 0;
    double ptail_coeff[6*TAILTERMS] = {0.0}, ntail_coeff[6*TAILTERMS] = {0.0};
    double dtmp = 1.0, expo = 0.0, erfc_factor, leading_exp;
    dd_t ptail[6], ntail[6], sum[6];
    dd_t meansquared, mu2, mu3, mu4, mu2sq;
    static double aold = -12349678.0, bold = -2757132.0;
    static double oldmom[6];
    static double oldmean = -5.0, oldvar, oldskew, oldkurt, oldlogfactor;
    static double olddskewda, olddskewdb, olddkurtda, olddkurtdb;

    /* Check for nans */
    if (!(a == a))
    {
	fprintf(stderr, "ERROR in johnint: a is %g\n", a);
	fflush(stderr);
    }
    if (!(b == b))
    {
	fprintf(stderr, "ERROR in johnint: b is %g\n", b);
	fflush(stderr);
    }

    /* Check old values */
    if (a == aold && b == bold && oldmean >= 0.0)
	goto set_result;

    aold = a;
    bold = b;

    /* Convert the arguments to the convenient range */
    if (a < 0.0)
	a = -a;
    if (b < 0.0)
    {
	b_is_negative = 1;
	b = -b;
    }

    /* Initialize static data */
    if (!initialized)
        john_init();

    /* Process the trivial case of a == 0 */
    if (a == 0.0)
    {
        dtmp = erfc(b)/2.0;
	for (j=0; j<6; ++j)
	{
	    sum[j] = dd_real(dtmp, 0.0);
	    oldmom[j] = dtmp;
	}
    }
    else
    {
        large_ratio = (b/a > LARGE_RATIO_BOUND &&
                       a >= LARGE_RATIO_MIN_A);

	/* Evaluate positive side tails */
	for (npos=0; npos<TAILTERMS; ++npos)
	{
            erfc_factor = erfc_exp_factor(a + b + (double)npos/2.0/a);
            leading_exp = -((a + b)*(a + b) + (double)npos);
            if (large_ratio)
            {
                for (j=0; j<6; ++j)
                {
                    expo = leading_exp + (double)(j+1)*b/a;
                    assert(expo < max_exponent);
                    dtmp = exp(expo)*erfc_factor;
                    ptail_coeff[j*TAILTERMS+npos] = 
                        dtmp*(double)pos_tail_coeff[j*TAILTERMS+npos];
                }
                if (expo < -max_exponent)
                    break;
                if (dtmp < DBL_MIN)
                    break;
            }
            else
            {
                if (leading_exp < -max_exponent)
                    break;
                dtmp = exp(leading_exp)*erfc_factor;
                if (dtmp < DBL_MIN)
                    break;
                for (j=0; j<6; ++j)
                    ptail_coeff[j*TAILTERMS+npos] = 
                        dtmp*(double)pos_tail_coeff[j*TAILTERMS+npos];
            }
	}
	for (j=0; j<6; ++j)
	    ptail[j] = dd_times_d(kahan_sum(ptail_coeff+j*TAILTERMS, npos), 0.5);

/*          printf("\nnpos = %d, pos tails are %g, %g, %g, %g, %g, %g\n", */
/*                 npos, ptail[0].hi, ptail[1].hi, ptail[2].hi, */
/*                 ptail[3].hi, ptail[4].hi, ptail[5].hi); */

	/* Evaluate negative side tails */
	for (nneg=1; nneg<TAILTERMS; ++nneg)
	{
            int lastcoeff = 0;
	    const double di = (double)nneg/2.0/a;
            const double erfc_arg = -b + a + di;
	    if (erfc_arg >= 0.0)
            {
                erfc_factor = erfc_exp_factor(erfc_arg);
                leading_exp = -((a - b)*(a - b) + (double)nneg);
            }
            else
            {
                erfc_factor = erfc(erfc_arg);
                leading_exp = di*di - (double)nneg*b/a;
            }
            if (large_ratio)
            {
                for (j=0; j<6; ++j)
                {
                    const int index = j*TAILTERMS+nneg;
                    if (neg_tail_coeff[index])
                    {
                        lastcoeff = neg_tail_coeff[index];                        
                        expo = leading_exp + (double)(j+1)*b/a;
                        dtmp = exp(expo)*erfc_factor;
                        ntail_coeff[index] = dtmp*(double)lastcoeff;
                    }
                }
            }
            else
            {
                expo = leading_exp;
                dtmp = exp(expo)*erfc_factor;
                for (j=0; j<6; ++j)
                {
                    const int index = j*TAILTERMS+nneg;
                    if (neg_tail_coeff[index])
                    {
                        lastcoeff = neg_tail_coeff[index];
                        ntail_coeff[index] = dtmp*(double)lastcoeff;
                    }
                }
            }
            if (lastcoeff)
            {
                if (expo < -max_exponent || dtmp < DBL_MIN)
                    break;
                if (expo > max_exponent || dtmp > DBL_MAX/fabs(lastcoeff))
                {
                    printf("WARNING in johnint: insufficient dynamic range\n");
                    break;
                }
            }
	}
	prec = dd_getprecision();
	dd_setprecision(DD_DBLPREC);
	for (j=0; j<6; ++j)
	    ntail[j] = dd_times_d(kahan_sum(ntail_coeff+j*TAILTERMS, nneg), 0.5);
	dd_setprecision(prec);

/*         printf("nneg = %d, neg tails are %g, %g, %g, %g, %g, %g\n", */
/*                nneg, ntail[0].hi, ntail[1].hi, ntail[2].hi, */
/*                ntail[3].hi, ntail[4].hi, ntail[5].hi); */

	/* Evaluate central parts of the integrals.
	   Use different formulas for large and small a. */
	if (a < SMALL_A_BOUND)
	    central_integ_small_a(a, b, sum);
	else
	    central_integ_big_a(a, b, sum);
        if (large_ratio)
            for (j=0; j<6; ++j)
            {
                if ((double)(j+1)*b/a < max_exponent)
                    sum[j] = dd_times_d(sum[j],exp((double)(j+1)*b/a));
                else if (sum[j].hi != 0.0)
                {
                    fprintf(stderr, "Function johnint: algorithm breakdown "
                            "at a = %g, b = %g. Aborting.\n", a, b);
                    fflush(stderr);
                    assert(0);
                }
            }

/*         printf("centrals are %g, %g, %g, %g, %g, %g\n", */
/*                sum[0].hi, sum[1].hi, sum[2].hi, */
/*                sum[3].hi, sum[4].hi, sum[5].hi); */

	/* Sum up the terms */
	prec = dd_getprecision();
	dd_setprecision(DD_DBLPREC);
	for (j=0; j<6; ++j)
	{
	    sum[j] = dd_sum(sum[j], dd_sum(ptail[j], ntail[j]));
	    oldmom[j] = sum[j].hi;
	}
	dd_setprecision(prec);
    }

    /* Calculate moments around the center */
    prec = dd_getprecision();
    dd_setprecision(DD_DBLPREC);

    meansquared = dd_square(sum[0]);
    mu2 = dd_diff(sum[1], meansquared);
    positive_variance = (mu2.hi > 0.0);
    mu3 = dd_diff(sum[2], dd_prod(dd_diff(dd_times_d(sum[1],3.0),
                                          dd_times_d(meansquared,2.0)),sum[0]));
    mu4 = dd_sum(dd_diff(sum[3], dd_times_d(dd_prod(sum[2],sum[0]),4.0)),
                 dd_times_d(dd_prod(meansquared,dd_sum(sum[1],mu2)),3.0));
    mu2sq = dd_square(mu2);

    oldmean = sum[0].hi;
    oldvar = mu2.hi;
    if (large_ratio)
        oldlogfactor = -b/a;
    else
        oldlogfactor = 0.0;
    if (positive_variance)
    {
        oldskew = sqrt(dd_div(dd_square(mu3),dd_prod(mu2sq, mu2)).hi);
        if (b_is_negative) oldskew *= -1.0;
        oldkurt = dd_div(mu4,mu2sq).hi;
    }

    dd_setprecision(prec);
    if (large_ratio)
        assert(positive_variance);

    /* Precise partial derivatives of skewness and kurtosis.
     * Use of quad precision for this calculation seems to be
     * unnecessary since it is not a dominant source of errors.
     */
    if (0 && !large_ratio)
    {
        double dkurta, dkurtb, dskewa, dskewb, skewdenom;
        dd_t deriv[4];

        skewdenom = pow(mu2.hi,2.5);
        prec = dd_getprecision();
        dd_setprecision(DD_DBLPREC);

        central_moment_partial_over_a(a, b, sum, deriv);
        dkurta = dd_diff(deriv[3],dd_times_d(
            dd_prod(deriv[1],dd_div(mu4,mu2)),2.0)).hi/mu2sq.hi;
        dskewa = dd_diff(dd_prod(deriv[2],mu2),dd_times_d(
            dd_prod(mu3,deriv[1]),1.5)).hi/skewdenom;

        central_moment_partial_over_b(a, b, sum, deriv);
        dkurtb = dd_diff(deriv[3],dd_times_d(
            dd_prod(deriv[1],dd_div(mu4,mu2)),2.0)).hi/mu2sq.hi;
        dskewb = dd_diff(dd_prod(deriv[2],mu2),dd_times_d(
            dd_prod(mu3,deriv[1]),1.5)).hi/skewdenom;

        dd_setprecision(prec);
        if (b_is_negative)
        {
            dkurtb *= -1.0;
            dskewa *= -1.0;
        }
        printf("dkurtda = %g, dkurtdb = %g, dskewda = %g, dskewdb = %g\n",
               dkurta, dkurtb, dskewa, dskewb);
    }

    /* Calculate partial derivatives of skewness and kurtosis */
    {
        const double ONE = 1.0, ONE5 = 1.5, TWO = 2.0;
        const double THREE = 3.0, FOUR = 4.0, SIX = 6.0;
        int J, K;
        double T, S, Y, H2, H2A, H2B, H3, H4, W, U, G, D;
        double *HMU, DERIV[5], DD[5];

        G = b*SQR2;
        D = a*SQR2;
        HMU = oldmom-1;
        H2  = oldvar;
        H2A = sqrt(H2) * H2;
        H2B = H2 * H2;
        H3 = fabs(oldskew) * H2A;
        H4 = oldkurt * H2B;
        W = G * D;
        U = D * D;
        for (J=1; J<=2; ++J)
        {
            for (K=1; K<=4; ++K)
            {
                T = (double)K;
                if (J == 1)
                    S = HMU[K + 1] - HMU[K];
                else
                    S = ((W - T) * (HMU[K] - HMU[K + 1]) + (T + ONE) *
                         (HMU[K + 1] - HMU[K + 2])) / U;
                DD[K] = T * S / D;
            }
            T = TWO * HMU[1] * DD[1];
            S = HMU[1] * DD[2];
            Y = DD[2] - T;
            DERIV[J] = (DD[3] - THREE * (S + HMU[2] * DD[1] - T * HMU[1])
                        - ONE5 * H3 * Y / H2) / H2A;
            DERIV[J + 2] = (DD[4] - FOUR * (DD[3] * HMU[1] + DD[1] * HMU[3])
                            + SIX * (HMU[2] * T + HMU[1] * (S - T * HMU[1]))
                            - TWO * H4 * Y / H2) / H2B;
        }
        olddskewdb = DERIV[1]*SQR2;
        olddkurtdb = DERIV[3]*SQR2;

        if (large_ratio)
        {
            double alpha, c0, c1, c2, ab;
            double I1, I2, I3, I4, I5, I6;

            alpha = exp(oldlogfactor);
            ab = a*b;
            I1 = oldmom[0];
            I2 = oldmom[1];
            I3 = oldmom[2];
            I4 = oldmom[3];
            I5 = oldmom[4];
            I6 = oldmom[5];

            c0 = (I1*(I1*I3 - I2*I2) - I3*oldvar);
            c1 = (-3 + 2*ab)*I2*I2*I2 + (-5 + 2*ab)*I3*I3 + 
                (-7 + 2*ab)*I1*I1*I4 + I2*(-4*(-2 + ab)*I1*I3 + 
                                           (7 - 2*ab)*I4);
            c2 = 2*I2*I2*I3 - 2*I1*I3*I3 - 3*I1*I2*I4 + 3*I3*I4 - 4*I5*oldvar;
            olddskewda = 3.0*(c2*alpha*alpha + c1*alpha + c0)/
                (a*a*a*oldvar*oldvar*sqrt(oldvar))/2.0;

            c0 = 6*I1*I1*I1*I3 - 2*I1*I2*I3 + 2*I2*I4 - 3*I1*I1*(I2*I2 + I4);
            c1 = ((3 - 2*ab)*I2*I2*I3 + 3*(-7 + 2*ab)*I1*I1*I1*I4 + 
                  (5 - 2*ab)*I3*I4 + I1*(3*(-3 + 2*ab)*I2*I2*I2 + 
                                         4*(-5 + 2*ab)*I3*I3 + (18 - 4*ab)*I2*I4) + 
                  (-9 + 2*ab)*I2*I5 + I1*I1*(-12*(-2 + ab)*I2*I3 + 
                                             (9 - 2*ab)*I5));
            c2 = (-2*I2*I3*I3 - 3*I4*I4 - 12*I1*I5*oldvar +
                  2*I1*(3*I2*I2*I3 + 7*I3*I4) + 
                  5*I6*oldvar - I1*I1*(6*I3*I3 + 9*I2*I4));
            olddkurtda = -2.0*(c2*alpha*alpha + c1*alpha + c0)/
                (a*a*a*oldvar*oldvar*oldvar);
        }
        else
        {
            olddskewda = DERIV[2]*SQR2;
            olddkurtda = DERIV[4]*SQR2;
        }

        if (b_is_negative)
        {
            olddkurtdb *= -1.0;
            olddskewda *= -1.0;
        }
    }

 set_result:
    if (p)
        for (j=0; j<6; ++j)
            p[j] = oldmom[j];
    if (mean)
    {
        assert(logfactor);
        *mean = oldmean;
    }
    if (var)
    {
        assert(logfactor);
        *var = oldvar;
    }
    if (skew) *skew = oldskew;
    if (kurt) *kurt = oldkurt;
    if (logfactor) *logfactor = oldlogfactor;
    if (dskewda) *dskewda = olddskewda;
    if (dskewdb) *dskewdb = olddskewdb;
    if (dkurtda) *dkurtda = olddkurtda;
    if (dkurtdb) *dkurtdb = olddkurtdb;
}

#ifdef TEST_SBSHAPE
#define EPS 1.0e-5

/* Functions from johnson.f */
double highmean_(const int *DEG, const double *Q, const double *W);
double highvar_(const int *DEG, const double *Q, const double *W);
void highb1_(const int *DEG, const double *Q, const double *W,
	     double *B1, double *DB1DQ, double *DB1DW);
void highexc_(const int *DEG, const double *Q, const double *W,
	      double *K, double *DKDQ, double *DKDW);

static double rd(double a, double b)
{
    if (a == 0.0 && b == 0.0)
	return 0.0;
    else
	return 2.0*fabs(a-b)/(fabs(a)+fabs(b));
}

static void big_b_moments(double a, double b, int deg,
			  int *ifail, double *mean, double *var,
			  double *skew, double *kurt)
{
    int b_is_negative = 0;
    const double min_ratio = 4.60517; /* log(100.0) */
    double q, w, b1, db1dq, db1dw, k, dkdq, dkdw;

    assert(ifail);
    *ifail = 1;
    if (a < 0.0)
	a = -a;
    if (b < 0.0)
    {
	b_is_negative = 1;
	b = -b;
    }
    /* The following requirement prevents overflow
       in calculating the expansion coefficients */
    if (a < 0.3)
	return;
    /* The following requirement ensures that
       the precision of the answer is satisfactory */
    if (b < min_ratio*a || b < a + 6.0/a + 6.0)
	return;

    q = exp(-b/a);
    w = exp(0.5/a/a);
    if (mean)
	*mean = highmean_(&deg, &q, &w);
    if (var)
	*var = highvar_(&deg, &q, &w);
    if (skew)
    {
	highb1_(&deg, &q, &w, &b1, &db1dq, &db1dw);
	*skew = sqrt(b1);
	if (b_is_negative)
	    *skew *= -1.0;
    }
    if (kurt)
    {
	highexc_(&deg, &q, &w, &k, &dkdq, &dkdw);
	*kurt = k+3.0;
    }
    *ifail = 0;
}

static void full_integ_big_a(const double a, const double b, dd_t *p)
{
    static double x[GAUSS_QUAD_POINTS];
    static int firstcall = 1;
    int i, j, prec;
    double expo, f;

    if (!initialized)
        john_init();

    if (firstcall)
    {
	double z;
	for (i=0; i<GAUSS_QUAD_POINTS; ++i)
	{
	    z = (gauss_abscissae[i] + 1.0)*0.5;
	    assert(z > 0.0 && z < 1.0);
	    x[i] = dgausn_(&z)/SQR2;
	}
	firstcall = 0;
    }

    for (i=0; i<GAUSS_QUAD_POINTS; ++i)
    {
	expo = gauss_weights[i];
	f = 1.0/(1.0+exp(-(x[i]-b)/a));
	for (j=0; j<6; ++j)
	{
	    expo *= f;
	    /* Avoid NaNs */
	    if (expo != 0.0)
		gauss_data[j*GAUSS_QUAD_POINTS+i] = expo;
	    else
		gauss_data[j*GAUSS_QUAD_POINTS+i] = 0.0;
	}
    }

    prec = dd_getprecision();
    dd_setprecision(DD_DBLPREC);
    for (j=0; j<6; ++j)
	p[j] = dd_times_d(kahan_sum(gauss_data+j*GAUSS_QUAD_POINTS,
				    GAUSS_QUAD_POINTS), 0.5);
    dd_setprecision(prec);
}

int main(int argc, char* argv[])
{
    /* This program tests recursion relations for integrals
       needed to evaluate parameters of the Johnson function */
    int n, ifail, arg_error = 0;
    double a, b, m_minus[7], m[7], m_plus[7];
    double mean, var, skew, kurt, mean2, var2, skew2, kurt2;
    double meanp, varp, skewp, kurtp, meanm, varm, skewm, kurtm;
    double dskewda, dskewdb, dkurtda, dkurtdb;
    double logfact, logcoeff;
    dd_t fullint[6];
    char *endptr;

    if (argc != 3)
	arg_error = 1;
    if (!arg_error)
    {
	a = strtod(argv[1], &endptr);
	if (endptr == argv[1])
	    arg_error = 1;
    }
    if (!arg_error)
    {
	b = strtod(argv[2], &endptr);
	if (endptr == argv[2])
	    arg_error = 1;
    }
    if (arg_error)
    {
	printf("\nUsage: %s a b  (a and b are doubles)\n\n", argv[0]);
	exit(EXIT_FAILURE);
    }
    
    johnint(a, b, m+1, &mean, &var, &skew, &kurt, &logfact,
            &dskewda, &dskewdb, &dkurtda, &dkurtdb);
    logcoeff = exp(logfact);
    mean    *= logcoeff;
    var     *= (logcoeff*logcoeff);
    dskewdb *= logcoeff;
    dkurtdb *= logcoeff;
    full_integ_big_a(a, b, fullint);
    printf("a = %g, b = %g, integrals are\n"
	   "johnin: %g %g %g %g %g %g\n",
           a, b, m[1], m[2], m[3], m[4], m[5], m[6]);
    printf("simple: %g %g %g %g %g %g\n",
	   fullint[0].hi, fullint[1].hi, fullint[2].hi,
	   fullint[3].hi, fullint[4].hi, fullint[5].hi);
    printf("\nNormal mean, var, skew, kurt = %g, %g, %g, %g\n",
           mean, var, skew, kurt);

    big_b_moments(a, b, 100, &ifail, &mean2, &var2, &skew2, &kurt2);
    if (ifail == 0) {
	printf("Big b  mean, var, skew, kurt = %g, %g, %g, %g\n",
	       mean2, var2, skew2, kurt2);
	printf("Relative differences are       %g, %g, %g, %g\n",
	       rd(mean, mean2), rd(var, var2),
	       rd(skew, skew2), rd(kurt, kurt2));
    }
    
    {
        int ifail;
        const size_t big_a_deg = 18;
        const double aderfactor  = -2.0/a/a/a;
        const double absb = fabs(b);
        double bderfactor = 0.0, crossfactor = 0.0;
        double bamean, bastdev, baskew, bakurt;
        double dbaskewdb, dbakurtdb, dbaskewda, dbakurtda;

        big_a_shape(a, b, big_a_deg, &ifail, &bamean, &bastdev,
                    &baskew, &bakurt, &dbaskewda,
                    &dbaskewdb, &dbakurtda, &dbakurtdb);
        if (ifail == 0)
        {
            /* Convert from derivatives over o = 1/a^2 and
             * f0 = 1/(1 + exp(absb/a)) into derivatives over a and b.
             */
            if (absb/a < max_exponent)
            {
                const double eba = exp(absb/a);
                bderfactor  = -eba/a/(1.0 + eba)/(1.0 + eba);
                crossfactor = -absb/a*bderfactor;
            }
            dbaskewda = dbaskewda*aderfactor + dbaskewdb*crossfactor;
            dbaskewdb *= bderfactor;
            dbakurtda = dbakurtda*aderfactor + dbakurtdb*crossfactor;
            dbakurtdb *= bderfactor;

            if (b == 0.0)
            {
                baskew = 0.0;
                dbakurtdb = 0.0;
                dbaskewda = 0.0;
            }
            else if (b < 0.0)
            {
                baskew *= -1.0;
                dbakurtdb *= -1.0;
                dbaskewda *= -1.0;
            }

            printf("Big a  mean, var, skew, kurt = %g, %g, %g, %g\n",
                   bamean, bastdev, baskew, bakurt);
            printf("Big a dskewdb = %g, dkurtdb = %g\n", dbaskewdb, dbakurtdb);
            printf("Big a dskewda = %g, dkurtda = %g\n", dbaskewda, dbakurtda);
        }
    }

    {
        double as, bs, jmean, jvar, jskew, jkurt, jdskewddelta;
        double jdskewdgamma, jdkurtddelta, jdkurtdgamma;

        as = a*SQR2;
        bs = b*SQR2;
        johnintf_(&bs, &as, &jmean, &jvar, &jskew, &jkurt,
                  &jdskewddelta, &jdskewdgamma, &jdkurtddelta,
                  &jdkurtdgamma);

        printf("johnintf mean, var, skew, kurt = %g, %g, %g, %g\n",
               jmean, jvar, jskew, jkurt);
        printf("johnintf dskewdb = %g, dkurtdb = %g\n",
               jdskewdgamma*SQR2, jdkurtdgamma*SQR2);
        printf("johnintf dskewda = %g, dkurtda = %g\n",
               jdskewddelta*SQR2, jdkurtddelta*SQR2);
    }

    /* Partial derivative over b */
    johnint(a, b-EPS, m_minus+1, &meanm, &varm,
            &skewm, &kurtm, &logfact, 0, 0, 0, 0);
    logcoeff = exp(logfact);
    meanm *= logcoeff;
    varm  *= (logcoeff*logcoeff);
    johnint(a, b+EPS, m_plus+1, &meanp, &varp,
            &skewp, &kurtp, &logfact, 0, 0, 0, 0);
    logcoeff = exp(logfact);
    meanp *= logcoeff;
    varp  *= (logcoeff*logcoeff);

    printf("\n");
    for (n=1; n<5; ++n)
        printf("n = %d, analyt. derivative over b is %g, numer. is %g\n",
               n, (double)(-n/a*(m[n]-m[n+1])),
               (double)((m_plus[n]-m_minus[n])/EPS/2.0));
    printf("Analytic dskewdb = %g, dkurtdb = %g\n", dskewdb, dkurtdb);
    printf("Numerical dskewdb = %g, dkurtdb = %g\n",
           (skewp-skewm)/EPS/2.0, (kurtp-kurtm)/EPS/2.0);

    /* Partial derivative over a */
    johnint(a-EPS, b, m_minus+1, &meanm, &varm,
            &skewm, &kurtm, &logfact, 0, 0, 0, 0);
    logcoeff = exp(logfact);
    meanm *= logcoeff;
    varm  *= (logcoeff*logcoeff);
    johnint(a+EPS, b, m_plus+1, &meanp, &varp,
            &skewp, &kurtp, &logfact, 0, 0, 0, 0);
    logcoeff = exp(logfact);
    meanp *= logcoeff;
    varp  *= (logcoeff*logcoeff);

    printf("\n");
    for (n=1; n<5; ++n)
        printf("n = %d, analyt. derivative over a is %g, numer. is %g\n",
               n, -n/a/a/2.0*((n/a-2.0*b)*m[n]+
                              (2.0*b-(2*n+1)/a)*m[n+1] + (n+1)/a*m[n+2]),
               (m_plus[n]-m_minus[n])/EPS/2.0);
    printf("Analytic dskewda = %g, dkurtda = %g\n", dskewda, dkurtda);
    printf("Numerical dskewda = %g, dkurtda = %g\n",
           (skewp-skewm)/EPS/2.0, (kurtp-kurtm)/EPS/2.0);

    return(0);
}
#endif
